def get_docsplitter_pipe(
    chunk_overlap=2,
    chunk_size=20,
    explode_splits=True,
    keep_separators=True,
    patterns_are_regex=False,
    split_patterns=["\n\n", "\n", " ", ""],
    trim_whitespace=True,
):
    from johnsnowlabs import nlp

    nlp.start()

    return nlp.LightPipeline(
        nlp.PipelineModel(
            stages=[
                nlp.DocumentAssembler().setInputCol("text"),
                nlp.DocumentCharacterTextSplitter()
                .setInputCols(["document"])
                .setOutputCol("splits")
                .setChunkSize(chunk_size)
                .setChunkOverlap(chunk_overlap)
                .setExplodeSplits(explode_splits)
                .setPatternsAreRegex(patterns_are_regex)
                .setSplitPatterns(split_patterns)
                .setTrimWhitespace(trim_whitespace)
                .setKeepSeparators(keep_separators),
            ]
        )
    )



from johnsnowlabs.abstract_base.base_enum import BaseEnum


class EMRClusterStates(BaseEnum):
    # https://docs.aws.amazon.com/emr/latest/ManagementGuide/emr-overview.html#emr-overview-cluster-lifecycle
    STARTING = "STARTING"
    BOOTSTRAPPING = "BOOTSTRAPPING"
    RUNNING = "RUNNING"
    WAITING = "WAITING"
    TERMINATING = "TERMINATING"
    TERMINATED = "TERMINATED"



from dataclasses import dataclass
from typing import Optional

from johnsnowlabs import settings
from johnsnowlabs.abstract_base.base_enum import BaseEnum
from johnsnowlabs.py_models.lib_version import LibVersion
from johnsnowlabs.py_models.primitive import LibVersionIdentifier


class DatabricksClusterStates(BaseEnum):
    # https://docs.databricks.com/dev-tools/api/latest/clusters.html#clusterclusterstate
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    RESTARTING = "RESTARTING"
    RESIZING = "RESIZING"
    TERMINATING = "TERMINATING"
    TERMINATED = "TERMINATED"
    ERROR = "ERROR"
    UNKNOWN = "UNKNOWN"


class JvmHardwareTarget(BaseEnum):
    gpu = "gpu"
    cpu = "cpu"
    m1 = "apple_silicon"
    aarch = "aarch"

    @classmethod
    def bool_choice_to_hardware(
        cls, gpu: bool = False, cpu: bool = False, m1: bool = False
    ) -> "JvmHardwareTarget":
        if gpu:
            return cls.gpu
        elif cpu:
            return cls.cpu
        elif m1:
            return cls.m1
        else:
            return cls.cpu

    @staticmethod
    def from_str(s):
        if s not in JvmHardwareTarget:
            bck = "\n"
            raise Exception(
                f"Invalid value for jvm_install_type: {s}"
                f" please specify on of:\n{bck.join([n.value for n in JvmHardwareTarget])}"
            )
        else:
            return JvmHardwareTarget(s)


class PyInstallTypes(BaseEnum):
    tar = "tar.gz"
    wheel = "whl"

    @staticmethod
    def from_str(s):
        bck = "\n"
        if s not in PyInstallTypes:
            raise Exception(
                f"Invalid value for py_install_type: {s}"
                f" please specify on of:\n{bck.join([n.value for n in PyInstallTypes])}"
            )
        else:
            return PyInstallTypes(s)


class SparkVersion(BaseEnum):
    # Broad versions
    spark3xx = LibVersion("3.x.x")
    spark31x = LibVersion("3.1.x")
    spark32x = LibVersion("3.2.x")
    spark33x = LibVersion("3.3.x")
    spark330 = LibVersion("3.3.0")
    spark322 = LibVersion("3.2.2")
    spark321 = LibVersion("3.2.1")
    spark320 = LibVersion("3.2.0")
    spark313 = LibVersion("3.1.3")
    spark312 = LibVersion("3.1.2")
    spark311 = LibVersion("3.1.1")
    spark303 = LibVersion("3.0.3")
    spark302 = LibVersion("3.0.2")
    spark301 = LibVersion("3.0.1")
    spark300 = LibVersion("3.0.0")


class LicenseType(BaseEnum):
    trial = "Trial"
    research = "Research"


class LicensePlattform(BaseEnum):
    none = None
    databricks = "Databricks"
    floating = "Floating"


class ProductName(BaseEnum):
    hc = "Spark-Healthcare"
    nlp = "Spark-NLP"
    ocr = "Spark-OCR"
    finance = "Spark-Finance"
    nlp_display = "NLP-Display"
    nlu = "nlu"
    jsl_full = "full"
    pyspark = "PySpark"
    spark = "spark"
    java = "java"
    python = "python"
    jsl_lib = "John Snow Labs Python Library"

    @staticmethod
    def from_str(s):
        bck = "\n"
        if s not in ProductName:
            raise Exception(
                f"Invalid Product to install: {s}"
                f" please specify on of:\n{bck.join([n.value for n in ProductName])}"
            )
        else:
            return ProductName(s)

    @staticmethod
    def from_jsl_api(s: str):
        if s == "Visual NLP":
            return ProductName.ocr
        if s == "Healthcare NLP":
            return ProductName.hc
        if s == "Spark NLP":
            return ProductName.hc
        if s == "Finance NLP":
            return ProductName.hc
        if s == "Legal NLP":
            return ProductName.hc

        raise ValueError(f"Unknown product name from jsl-api {s}")


class ProductLogo(BaseEnum):
    hc = "💊"  # 🏥 🩺 💊 ❤️ ‍🩹 ‍⚕️💉 🥼 🚑 🔬 🫀 🩻 🧪
    nlp = "🚀"
    ocr = "🕶"  # 👁️  🤖 🦾🦿 🥽 👀 🕶 🥽 ⚕
    finance = "🤑"  # 🤑🏦💲💳💰💸💵💴💶💷
    nlp_display = "🎨"
    nlu = "🤖"
    jsl_full = "💯🕴"  # 🕴
    java = "🫘"  # 🫘 # ☕ ♨ 🥃 🥃 🧋🍹♨️🥤🫖
    python = "🐍"  # 🐉
    pyspark = "🐍+⚡"
    spark = "⚡ "
    databricks = "🧱"
    jsl_lib = "🧪"


class ProductSlogan(BaseEnum):
    healthcare = "Heal the planet with NLP!"
    spark_nlp = "State of the art NLP at scale"
    ocr = "Empower your NLP with a set of eyes"
    pyspark = "The big data Engine"
    nlu = "1 line of code to conquer nlp!"
    jsl_full = "The entire John Snow Labs arsenal!"
    finance = "NLP for the Finance Industry"
    nlp_display = "Visualize and Explain NLP!"
    jsl_lib = "Easy access to all of John Snow Labs Enterprise Software!"
    spark = "⚡"
    java = "☕"
    python = "🐍"


@dataclass
class InstalledProductInfo:
    """Representation of a JSL product install. Version is None if not installed"""

    product: ProductName
    version: Optional[LibVersionIdentifier] = None


@dataclass
class JslSuiteStatus:
    """Representation and install status of all JSL products and its dependencies.
    Version attribute of InstalledProductInfo is None for uninstalled products
    """

    spark_nlp_info: Optional[InstalledProductInfo] = None
    spark_hc_info: Optional[InstalledProductInfo] = None
    spark_ocr_info: Optional[InstalledProductInfo] = None
    nlu_info: Optional[InstalledProductInfo] = None
    sparknlp_display_info: Optional[InstalledProductInfo] = None
    pyspark_info: Optional[InstalledProductInfo] = None


class LatestCompatibleProductVersion(BaseEnum):
    jsl_lib = LibVersion(settings.raw_version_jsl_lib)
    healthcare = LibVersion(settings.raw_version_medical)
    spark_nlp = LibVersion(settings.raw_version_nlp)
    ocr = LibVersion(settings.raw_version_ocr)
    nlu = LibVersion(settings.raw_version_nlu)
    nlp_display = LibVersion(settings.raw_version_nlp_display)
    pyspark = LibVersion(settings.raw_version_pyspark)
    finance = LibVersion("finance")
    spark = LibVersion(settings.raw_version_pyspark)
    java = LibVersion("java")
    python = LibVersion("python")

    @staticmethod
    def from_settings(p: ProductName):
        if p == ProductName.hc:
            return settings.raw_version_medical
        if p == ProductName.ocr:
            return settings.raw_version_ocr
        if p == ProductName.nlp:
            return settings.raw_version_nlp

    @staticmethod
    def sct_from_settings(p: ProductName):
        if p == ProductName.hc:
            return settings.raw_version_secret_medical
        if p == ProductName.ocr:
            return settings.raw_version_secret_ocr
        if p == ProductName.nlp:
            return settings.raw_version_nlp



import os
from typing import Dict, Optional, Union

from johnsnowlabs import settings
from johnsnowlabs.abstract_base.pydantic_model import WritableBaseModel

# from johnsnowlabs.abstract_base.software_product import AbstractSoftwareProduct
from johnsnowlabs.py_models.jsl_secrets import JslSecrets
from johnsnowlabs.py_models.lib_version import LibVersion
from johnsnowlabs.utils.enums import JvmHardwareTarget, ProductName, PyInstallTypes


class InstallFileInfoBase(WritableBaseModel):
    file_name: str
    product: ProductName
    compatible_spark_version: Union[str, LibVersion]
    product_version: Union[str, LibVersion]

    # install_type: Optional[JvmHardwareTarget]
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.compatible_spark_version = LibVersion(self.compatible_spark_version)
        self.product_version = LibVersion(self.product_version)


class PyInstallInfo(InstallFileInfoBase):
    install_type: PyInstallTypes


class JvmInstallInfo(InstallFileInfoBase):
    install_type: JvmHardwareTarget


class LocalPyLib(WritableBaseModel):
    py_lib: Optional[PyInstallInfo] = None

    def get_py_path(self):
        if not self.py_lib:
            return False
        if not os.path.exists(os.path.join(settings.py_dir, self.py_lib.file_name)):
            return False
        return os.path.join(settings.py_dir, self.py_lib.file_name)


class LocalPy4JLib(WritableBaseModel):
    java_lib: Optional[JvmInstallInfo] = None
    py_lib: Optional[PyInstallInfo] = None

    def get_java_path(self):
        if not self.java_lib:
            return False
        if not os.path.exists(f"{settings.java_dir}/{self.java_lib.file_name}"):
            return False
        return f"{settings.java_dir}/{self.java_lib.file_name}"

    def get_py_path(self):
        if not self.py_lib:
            return False
        if not os.path.exists(f"{settings.py_dir}/{self.py_lib.file_name}"):
            return False
        return f"{settings.py_dir}/{self.py_lib.file_name}"


class RootInfo(WritableBaseModel):
    version: Union[str, LibVersion]
    run_from: str

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.version = LibVersion(self.version)

    @staticmethod
    def get_from_jsl_home():
        return RootInfo.parse_file(settings.root_info_file)


class InstallFolder(WritableBaseModel):
    infos: Dict[str, Union[PyInstallInfo, JvmInstallInfo]]

    def get_product_entry(
        self,
        product: ProductName,
        hardware_target: Optional[Union[PyInstallTypes, JvmHardwareTarget]] = None,
    ):
        for file_name, install_info in self.infos.items():
            if install_info.product == product:
                if hardware_target:
                    if install_info.install_type == JvmHardwareTarget(hardware_target):
                        return install_info
                else:
                    return install_info

    @staticmethod
    def java_folder_from_home():
        if os.path.exists(settings.java_info_file):
            return InstallFolder.parse_file(settings.java_info_file)
        return False

    @staticmethod
    def py_folder_from_home():
        if os.path.exists(settings.py_info_file):
            return InstallFolder.parse_file(settings.py_info_file)
        return False


class InstallSuite(WritableBaseModel):
    info: Optional[RootInfo] = None
    secrets: Optional[JslSecrets] = None
    # Py4J Libs
    nlp: Optional[LocalPy4JLib] = None
    ocr: Optional[LocalPy4JLib] = None
    hc: Optional[LocalPy4JLib] = None
    # Pure Python Libs
    pure_py_jsl: Optional[LocalPyLib] = None

    @staticmethod
    def empty():
        return InstallSuite()

    def get_missing_products(self, nlp, visual, spark_nlp):
        missing = []
        from johnsnowlabs.auto_install.softwares import Software

        if self.secrets.OCR_LICENSE and visual:
            if not self.ocr.java_lib or not self.ocr.get_java_path():
                missing.append(Software.spark_ocr)
        if self.secrets.HC_LICENSE and nlp:
            if not self.hc.java_lib or not self.hc.get_java_path():
                missing.append(Software.spark_hc)
        if spark_nlp:
            if not self.nlp.java_lib or not self.nlp.get_java_path():
                missing.append(Software.spark_nlp)
        return missing

    def log_missing_jars(
        self,
        should_have_ocr,
        should_have_hc,
        should_have_nlp,
    ):
        print(f"🚨 Looks like some of the missing jars could not be fetched...")
        if not self.ocr.java_lib and self.secrets.OCR_LICENSE and should_have_ocr:
            print(f"🚨 Missing Jar for OCR")
        if not self.hc.java_lib and self.secrets.HC_LICENSE and should_have_hc:
            print(f"🚨 Missing Jar for Medical")
        if not self.nlp.java_lib and should_have_nlp:
            print(f"🚨 Missing Jar for NLP")



import sys

from johnsnowlabs import *
import unittest
import pkg_resources


import os

os.environ["PYSPARK_PYTHON"] = sys.executable
os.environ["PYSPARK_DRIVER_PYTHON"] = sys.executable


# finance.ClassifierDLApproach()
class ImportTestCase(unittest.TestCase):
    def test_sparknlp_session(self):
        nlp.start()
        d = nlp.DocumentAssembler().setInputCol("text").setOutputCol("doc")
        t = nlp.Tokenizer().setInputCols("doc").setOutputCol("tok")
        c = (
            nlp.DeBertaForTokenClassification()
            .setInputCols(["tok", "doc"])
            .setOutputCol("class")
        )
        p = nlp.Pipeline(stages=[d, t])
        p = nlp.to_nlu_pipe(p)
        print(p.predict("Hello World"))

    def test_sparknlp_gpu_session(self):
        nlp.start(hardware_target="gpu")
        d = nlp.DocumentAssembler().setInputCol("text").setOutputCol("doc")
        t = nlp.Tokenizer().setInputCols("doc").setOutputCol("tok")
        c = (
            nlp.DeBertaForTokenClassification()
            .setInputCols(["tok", "doc"])
            .setOutputCol("class")
        )
        p = nlp.Pipeline(stages=[d, t])
        p = nlp.to_nlu_pipe(p)
        print(p.predict("Hello form John SNow labs"))

    def test_sparknlp_m1_session(self):
        import os

        nlp.start(hardware_target="m1")
        d = nlp.DocumentAssembler().setInputCol("text").setOutputCol("doc")
        t = nlp.Tokenizer().setInputCols("doc").setOutputCol("tok")
        c = (
            nlp.DeBertaForTokenClassification()
            .pretrained()
            .setInputCols(["tok", "doc"])
            .setOutputCol("class")
        )
        nlp.UniversalSentenceEncoder.pretrained()
        p = nlp.Pipeline(stages=[d, t])
        p = nlp.to_nlu_pipe(p)
        print(p.predict("Hello form John SNow labs"))

    def test_healthcare_session(self):
        nlp.start()
        d = nlp.DocumentAssembler().setInputCol("text").setOutputCol("doc")
        t = nlp.Tokenizer().setInputCols("doc").setOutputCol("tok")
        c = (
            medical.BertForTokenClassification()
            .pretrained()
            .setInputCols(["tok", "doc"])
            .setOutputCol("class")
        )
        p = nlp.Pipeline(stages=[d, t, c])
        p = nlp.to_nlu_pipe(p)
        print(p.predict("Hello form John SNow labs"))

    def test_ocr_session(self):
        # Convert pdf to image
        p = "/home/ckl/old_home/ckl/Documents/freelance/johnsnowlabs_lib/tmp/licenses/4_1_LATEST_OCR_HC_BCK.json"
        spark = nlp.start(visual=True)

        pdf_to_image = visual.PdfToImage()
        pdf_to_image.setImageType(visual.ImageType.TYPE_3BYTE_BGR)

        # Detect tables on the page using pretrained model
        # It can be finetuned for have more accurate results for more specific documents
        table_detector = visual.ImageTableDetector.pretrained(
            "general_model_table_detection_v2", "en", "clinical/ocr"
        )
        table_detector.setInputCol("image")
        table_detector.setOutputCol("region")

        # Draw detected region's with table to the page
        draw_regions = visual.ImageDrawRegions()
        draw_regions.setInputCol("image")
        draw_regions.setInputRegionsCol("region")
        draw_regions.setOutputCol("image_with_regions")
        draw_regions.setRectColor(visual.Color.red)

        # Extract table regions to separate images
        splitter = visual.ImageSplitRegions()
        splitter.setInputCol("image")
        splitter.setInputRegionsCol("region")
        splitter.setOutputCol("table_image")
        splitter.setDropCols("image")

        # Detect cells on the table image
        cell_detector = visual.ImageTableCellDetector()
        cell_detector.setInputCol("table_image")
        cell_detector.setOutputCol("cells")
        cell_detector.setAlgoType("morphops")

        # Extract text from the detected cells
        table_recognition = visual.ImageCellsToTextTable()
        table_recognition.setInputCol("table_image")
        table_recognition.setCellsCol("cells")
        table_recognition.setMargin(3)
        table_recognition.setStrip(True)
        table_recognition.setOutputCol("table")

        pipeline = nlp.PipelineModel(
            stages=[
                pdf_to_image,
                table_detector,
                draw_regions,
                splitter,
                cell_detector,
                table_recognition,
            ]
        )

        import pkg_resources

        pdf_example = pkg_resources.resource_filename(
            "sparkocr", "resources/ocr/pdfs/tabular-pdf/data.pdf"
        )
        pdf_example_df = spark.read.format("binaryFile").load(pdf_example).cache()
        pipeline.transform(pdf_example_df).show()

    def test_legal_session(self):
        nlp.start()

        nlp.Pipeline(self.get_legal_pipe()).fullAnnotate("Shwrm")

    def test_finance_session(self):
        nlp.start()
        nlp.Pipeline(self.get_finance_pipe()).fullAnnotate("unit")

    @staticmethod
    def get_finance_pipe() -> nlp.PipelineModel:
        documentAssembler = (
            nlp.DocumentAssembler().setInputCol("text").setOutputCol("ner_chunk")
        )

        embeddings = (
            nlp.UniversalSentenceEncoder.pretrained("tfhub_use", "en")
            .setInputCols("ner_chunk")
            .setOutputCol("sentence_embeddings")
        )

        resolver = (
            finance.SentenceEntityResolverModel.pretrained(
                "finel_tickers2names", "en", "finance/models"
            )
            .setInputCols(["ner_chunk", "sentence_embeddings"])
            .setOutputCol("name")
            .setDistanceFunction("EUCLIDEAN")
        )

        return nlp.PipelineModel(stages=[documentAssembler, embeddings, resolver])

    @staticmethod
    def get_legal_pipe() -> nlp.PipelineModel:
        z = legal.ZeroShotRelationExtractionModel.pretrained(
            "finre_zero_shot", "en", "finance/models"
        )
        documentAssembler = (
            nlp.DocumentAssembler().setInputCol("text").setOutputCol("ner_chunk")
        )

        embeddings = (
            nlp.UniversalSentenceEncoder.pretrained("tfhub_use", "en")
            .setInputCols("ner_chunk")
            .setOutputCol("sentence_embeddings")
        )

        resolver = (
            legal.SentenceEntityResolverModel.pretrained(
                "legel_crunchbase_companynames", "en", "legal/models"
            )
            .setInputCols(["ner_chunk", "sentence_embeddings"])
            .setOutputCol("name")
            .setDistanceFunction("EUCLIDEAN")
        )

        return nlp.PipelineModel(stages=[documentAssembler, embeddings, resolver])

    @staticmethod
    def get_cross_lib_pipe() -> nlp.PipelineModel:
        # Returns pipe with one anno per lib
        # TODO add some fancy OCR DL models?
        doc2text = visual.DocToText().setInputCol("content").setOutputCol("text")
        d = nlp.DocumentAssembler().setInputCol("text").setOutputCol("doc")
        t = nlp.Tokenizer().setInputCols("doc").setOutputCol("tok")
        # One classifier per NLP lib

        c1 = (
            medical.BertForTokenClassifier()
            .pretrained()
            .setInputCols(["tok", "doc"])
            .setOutputCol("medical")
        )

        c2 = (
            nlp.DeBertaForTokenClassification()
            .setInputCols(["tok", "doc"])
            .setOutputCol("opene_source")
        )

        c3 = (
            finance.BertForSequenceClassification.pretrained(
                "finclf_augmented_esg", "en", "finance/models"
            )
            .setInputCols(["tok", "doc"])
            .setOutputCol("finance")
        )

        c4 = (
            legal.BertForSequenceClassification.pretrained(
                "legclf_bert_judgements_agent", "en", "legal/models"
            )
            .setInputCols(["tok", "doc"])
            .setOutputCol("legal")
        )

        return nlp.Pipeline(stages=[doc2text, d, t, c1, c2, c3, c4])

    def test_simple_cross_lib(self):
        spark = nlp.start()
        doc_example = pkg_resources.resource_filename(
            "sparkocr", "resources/ocr/docs/doc2.docx"
        )
        df = spark.read.format("binaryFile").load(doc_example).cache()
        self.get_cross_lib_pipe().fit(df).transform(df).show()

    def test_simple_cross_lib_gpu(self):
        spark = nlp.start(hardware_target="gpu")
        doc_example = pkg_resources.resource_filename(
            "sparkocr", "resources/ocr/docs/doc2.docx"
        )
        df = spark.read.format("binaryFile").load(doc_example).cache()
        self.get_cross_lib_pipe().fit(df).transform(df).show()

    def test_cross_engine_session(self):
        import itertools

        # Test every combination of jars with CPU jars
        for c in range(3):
            p = itertools.combinations(["nlp-cpu", "ocr", "hc"], c)
            for pp in p:
                print(pp)

        # Test every combination of jars with GPU jars
        for c in range(3):
            p = itertools.combinations(["nlp-gpu", "ocr", "hc"], c)
            for pp in p:
                print(pp)


if __name__ == "__main__":
    unittest.main()



from typing import List, Optional, Tuple

import botocore

from johnsnowlabs.py_models.install_info import InstallSuite, LocalPy4JLib
from johnsnowlabs.utils.boto_utils import BotoException
from johnsnowlabs.utils.s3_utils import create_bucket, upload_file_to_s3


def create_glue_bucket(boto_session: "boto3.Session", bucket=None):
    """Create a bucket for EMR cluster logs
    :param boto_session: Boto3 session
    :param bucket: Bucket name
    """
    import boto3

    try:
        sts_client = boto_session.client("sts")
        account_id = sts_client.get_caller_identity()["Account"]
        region = boto_session.region_name
        if not bucket:
            bucket = f"aws-glue-assets-{account_id}-{region}"

        return create_bucket(boto_session=boto_session, bucket=bucket)

    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "BucketAlreadyOwnedByYou":
            return bucket
        raise BotoException(
            code=e.response["Error"]["Code"], message=e.response["Error"]["Message"]
        )


def upload_pylibs_jars_to_glue_bucket(
    boto_session: "boto3.Session", install_suite: InstallSuite, bucket: Optional[str]
) -> Tuple[List[str], List[str]]:
    """Uploads jars and python packages to glue bucket
    :param boto_session: Boto3 session
    :param install_suite: Install suite
    :param bucket: Bucket name
    :return: List of uploaded jars and python packages
    """
    import boto3

    if not boto_session:
        raise ValueError("Boto3 session is required")

    libs: List[LocalPy4JLib] = [install_suite.nlp, install_suite.ocr, install_suite.hc]
    uploaded_python_packages = []
    uploaded_jars = []
    for lib in libs:
        if lib:
            if lib.java_lib:
                print(f"👷 Uploading {lib.java_lib.file_name} ...")
                s3_jar_file_name = upload_file_to_s3(
                    boto_session=boto_session,
                    file_path=lib.get_java_path(),
                    bucket=bucket,
                    file_name=f"assets/jars/{lib.java_lib.file_name}",
                )
                uploaded_jars.append(s3_jar_file_name)
                print(f"✅ Successfully uploaded to {s3_jar_file_name}")
            if lib.py_lib:
                print(f"👷 Uploading {lib.py_lib.file_name} ...")
                pypi_jar_file_name = upload_file_to_s3(
                    boto_session=boto_session,
                    file_path=lib.get_py_path(),
                    bucket=bucket,
                    file_name=f"assets/packages/{lib.py_lib.file_name}",
                )
                uploaded_python_packages.append(pypi_jar_file_name)
                print(f"✅ Successfully uploaded to {pypi_jar_file_name}")
    return [uploaded_jars, uploaded_python_packages]



from haystack import Document
from mlflow.models import ModelSignature, infer_signature
import os
from haystack.pipelines import RayPipeline

# https://docs.haystack.deepset.ai/docs/pipelines#distributed-pipelines-with-ray
from haystack.nodes import EmbeddingRetriever

from johnsnowlabs.frameworks.embedding_retrieval.haystack_node import (
    JohnSnowLabsHaystackProcessor,
)

# os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"
from johnsnowlabs.llm import embedding_retrieval


def get_docs():
    return [
        Document(
            content="I like apples",
            content_type="text",
            id=1,
        ),
        Document(
            content="I like bananas \n and other things \n like icream \n and cats",
            content_type="text",
            id=2,
        ),
    ]


def test_integration():
    from haystack.nodes import PreProcessor
    from haystack import Pipeline
    from haystack.document_stores import InMemoryDocumentStore

    # processor = PreProcessor(
    #     clean_whitespace=True,
    #     split_by="word",
    # )

    processor = JohnSnowLabsHaystackProcessor()

    # Write some processed data to Doc store, so we can retrieve it later
    document_store = InMemoryDocumentStore(embedding_dim=512)
    document_store.write_documents(processor.process(get_docs()))

    # could just use EmbeddingRetriever but mehhh
    retriever = embedding_retrieval.JohnSnowLabsHaystackEmbedder(
        embedding_model="en.embed_sentence.bert_base_uncased",
        # model_format="johnsnowlabs",
        document_store=document_store,
        use_gpu=False,
    )
    document_store.update_embeddings(retriever)

    pipe = Pipeline()
    # pipe.add_node(component=processor, name="Preprocess", inputs=["Query"])
    pipe.add_node(component=processor, name="Preprocess", inputs=["Query"])
    pipe.add_node(component=retriever, name="Embed&Retrieve", inputs=["Query"])
    result = pipe.run(documents=get_docs(), query="lol")
    print(result)



import uvicorn, os
from fastapi import FastAPI, HTTPException
from johnsnowlabs import *
import logging

# Managing logging level
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
logging.getLogger("py4j.java_gateway").setLevel(logging.ERROR)

# FastAPI object
app = FastAPI()

# Start a Spark NLP Session
nlp.start()

# Always preload your objects to reduce latency
preloaded = dict()
preloaded['legpipe_deid'] = nlp.PretrainedPipeline('legpipe_deid', 'en', 'legal/models')


# ENDPOINTS
@app.get("/legpipe_deid")
async def get_one_sequential_pipeline_result(text=None):
    if text is None:
        example = """Pizza Fusion Holdings, Inc. Franchise Agreement This Franchise Agreement (the "Agreement") is entered into as of the Agreement Date shown on the cover page between Pizza Fusion Holding, Inc., a Florida corporation, and the individual or legal entity identified on the cover page. Source: PF HOSPITALITY GROUP INC., 9/23/2015t."""
    else:
        example = text

    ann_result = preloaded["legpipe_deid"].fullAnnotate(example)

    result = {}
    for column, ann in ann_result[0].items():
        result[column] = []
        for lines in ann:
            content = {
                "result": lines.result,
                "begin": lines.begin,
                "end": lines.end,
                "metadata": dict(lines.metadata),
            }
            result[column].append(content)
    return result


# FASTAPI start
if __name__ == "__main__":
    uvicorn.run('fastapi_app:app', host='0.0.0.0', port=8515)



# get helpers into global space
from johnsnowlabs import nlp, settings, viz, lab
from .abstract_base.lib_resolver import try_import_lib

# These we may only import, if the corresponding package is installed
# Otherwise we will run modules __init__ to early and will not attach new classes/methods to it
if try_import_lib("sparknlp_jsl") and try_import_lib("sparknlp"):
    from johnsnowlabs import medical, finance, legal
    from sparknlp_jsl.alab import AnnotationLab

    lab = AnnotationLab()

if try_import_lib("sparkocr") and try_import_lib("sparknlp"):
    from johnsnowlabs import visual

from johnsnowlabs import llm


def new_version_online():
    from .utils.pip_utils import get_latest_lib_version_on_pypi

    # we are outdated, if current version does not match the latest on PypPi
    from .auto_install.softwares import Software

    return settings.raw_version_jsl_lib != get_latest_lib_version_on_pypi(
        Software.jsl_lib.pypi_name
    )



from johnsnowlabs.auto_install.databricks.endpoints import *
from databricks.sdk.runtime import *


def log_response_if_bad_status(response):
    try:
        response.raise_for_status()
    except Exception as err:
        print(response.text)
        raise err


from johnsnowlabs.auto_install.databricks.endpoints import *


def create_endpoint(
    endpoint_name,
    model_name,
    model_version,
    secret_scope_name="JSL_SCOPE",
    secret_name="JSL_SECRET_NAME",
    workload_size="Small",
    block_until_deployed=True,
    workload_type="CPU",
    db_token=None,
    db_host=None,
):
    """Create serving endpoint and wait for it to be ready
    maps name of your secret to an env variable with the same name in the container
    """

    print(f"Creating new serving endpoint: {endpoint_name}")
    endpoint_url = f"{db_host}/api/2.0/serving-endpoints"
    served_models = [
        {
            "name": "current",
            "model_name": model_name,
            "model_version": model_version,
            "workload_size": workload_size,
            "workload_type": workload_type,
            "scale_to_zero_enabled": "false"
            if "gpu" in workload_type.lower()
            else "true",
            "env_vars": [
                {
                    "env_var_name": "JOHNSNOWLABS_LICENSE_JSON",
                    "secret_scope": secret_scope_name,
                    "secret_key": secret_name,
                },
                {
                    "env_var_name": "spark.databricks.api.url",
                    "secret_scope": secret_scope_name,
                    "secret_key": "DB_API_URL",
                },
                {
                    "env_var_name": "DB_ENDPOINT_ENV",
                    "secret_scope": secret_scope_name,
                    "secret_key": "DB_ENDPOINT_ENV",
                },
            ],
        }
    ]

    request_data = {"name": endpoint_name, "config": {"served_models": served_models}}
    response = requests.post(
        endpoint_url,
        data=json.dumps(request_data).encode("utf-8"),
        headers=get_headers(db_token),
    )
    response_json = response.json()
    if (
        "error_code" in response_json
        and response_json["error_code"] == "RESOURCE_DOES_NOT_EXIST"
    ):
        print("Model not imported from marketplace")
        return "RESOURCE_DOES_NOT_EXIST"
    elif (
        "error_code" in response_json
        and response_json["error_code"] == "RESOURCE_ALREADY_EXISTS"
    ):
        return "RESOURCE_ALREADY_EXISTS"

    log_response_if_bad_status(response)
    if block_until_deployed:
        wait_for_endpoint(endpoint_name, db_host, db_token)

    try:
        displayHTML(
            f"""Created the <a href="/#mlflow/endpoints/{endpoint_name}" target="_blank">{endpoint_name}</a> serving endpoint"""
        )
    except:
        print(
            f"Created serving endpoint {endpoint_name} at {db_host}/#mlflow/endpoints/{endpoint_name}"
        )
    return True


import os

from johnsnowlabs.auto_install.databricks.endpoints import (
    setup_secrets,
)


def make_model_select_drop_down(models_df):
    def extract_number(s):
        import re

        # Use regular expression to extract the numeric part
        match = re.search(r"\d+", s)
        return int(match.group()) if match else 0

    models = sorted(models_df.DropDownId.values.tolist(), key=extract_number)
    first_model = models[0]
    dbutils.widgets.dropdown("The model", first_model, models)


def get_selected_widget_model_metadata(models_df):
    selected_model = dbutils.widgets.get("The model")
    model_data = models_df[models_df.DropDownId == selected_model]
    return model_data


def get_model_metadta(models_df, selected_model):
    model_data = models_df[models_df.DropDownId == selected_model]
    return model_data


def query_endpoint(data, endpoint_name, db_host, db_token, base_name=None):
    url = f"{db_host}/serving-endpoints/{endpoint_name}/invocations"
    headers = {
        "Authorization": f"Bearer {db_token}",
        "Content-Type": "application/json",
    }
    response = requests.request(method="POST", headers=headers, url=url, data=data)
    if response.status_code != 200:
        raise Exception(
            f"Request failed with status {response.status_code}, {response.text}"
        )
    import pandas as pd

    return pd.DataFrame(json.loads(response.json()["predictions"]))


def render_ui():
    from johnsnowlabs.auto_install.databricks.marketplace_offering import models_df

    dbutils.widgets.removeAll()
    dbutils.widgets.text("JSL-License", "")
    dbutils.widgets.text("Databricks access token", "")
    dbutils.widgets.text("Databricks host", "")
    make_model_select_drop_down(models_df)
    dbutils.widgets.dropdown("hardware_target", "CPU", ["CPU", "GPU"])


def get_db_token():
    return dbutils.widgets.get("Databricks access token")


def get_db_host():
    return dbutils.widgets.get("Databricks host")


def get_hardware_target():
    return dbutils.widgets.get("hardware_target")


def get_jsl_license():
    return dbutils.widgets.get("JSL-License")


def deploy(deployed_endpoint_name=None, jsl_model_id=None):
    from johnsnowlabs.auto_install.databricks.marketplace_offering import models_df

    db_token = get_db_token()
    db_host = get_db_host()
    os.environ["DATABRICKS_HOST"] = db_host
    os.environ["DATABRICKS_TOKEN"] = db_token
    hardware_target = get_hardware_target()
    jsl_license = get_jsl_license()

    jsl_license = json.dumps({"SPARK_NLP_LICENSE": jsl_license})
    if not jsl_model_id:
        model_data = get_selected_widget_model_metadata(models_df)
    else:
        model_data = get_model_metadta(models_df, jsl_model_id)

    model_path = (
        model_data.CpuModelPath.values[0]
        if hardware_target == "CPU"
        else model_data.GpuModelPath.values[0]
    )
    endpoint_name = (
        (model_path + "_endpoint")[:50].replace(".", "") + "_" + hardware_target.lower()
        if not deployed_endpoint_name
        else deployed_endpoint_name
    )

    if len(endpoint_name) > 60:
        print("Endpoint name is too long, truncating to 60 characters")
        endpoint_name = endpoint_name[:60]

    setup_secrets(
        secret_name="JSL_SECRET_NAME",
        secret_value=jsl_license,
        scope_name="JSL_SCOPE",
        host=db_host,
        db_token=db_token,
    )

    print("creating model", model_path)
    print("creating endpoint", endpoint_name)

    endpoint_success = create_endpoint(
        endpoint_name,
        model_path,
        "2",
        db_token=db_token,
        db_host=db_host,
        workload_type=hardware_target,
    )

    if endpoint_success == "RESOURCE_DOES_NOT_EXIST":
        listing_id = model_data.ListingId.values[0]
        try:
            displayHTML(
                f"""Could not import the model. <a href="marketplace/consumer/listings/{listing_id}" target="_blank">Please click this link and click on "get instant access" in the top right</a> """
            )
        except:
            print(
                f"""Could not import the model. Please visit  {db_host}/marketplace/consumer/listings/{listing_id} and click on "get instant access" in the top righ"""
            )
        return False
    elif endpoint_success == "RESOURCE_ALREADY_EXISTS":
        print(
            f"Endpoint with name {endpoint_name} already exists. Provide a different name or deploy a different model. "
        )
        return False
    return endpoint_name



import pytest

from johnsnowlabs.frameworks.embedding_retrieval import JohnSnowLabsCharSplitter


# https://colab.research.google.com/drive/1J7JpxIkYcOlm01otttJLD6iFLge43dZ9?usp=sharing


def test_integration():
    from langchain.document_loaders import TextLoader
    from langchain.text_splitter import CharacterTextSplitter
    from langchain.vectorstores import FAISS
    from langchain.embeddings import OpenAIEmbeddings
    from langchain.agents.agent_toolkits import create_retriever_tool

    from johnsnowlabs.llm import embedding_retrival

    p = "/home/ckl/Documents/freelance/jsl/johnsnowlabs-4-real/tests/datasets/state_of_the_union.txt"
    loader = TextLoader(p)
    documents = loader.load()
    text_splitter = CharacterTextSplitter(chunk_size=1000, chunk_overlap=0)
    texts = text_splitter.split_documents(documents)
    embeddings = embedding_retrival.JohnSnowLabsLangChainEmbedder(
        "en.embed_sentence.bert_base_uncased"
    )  #  OpenAIEmbeddings()
    db = FAISS.from_documents(texts, embeddings)
    retriever = db.as_retriever()
    tool = create_retriever_tool(
        retriever,
        "search_state_of_union",
        "Searches and returns documents regarding the state-of-the-union.",
    )
    tools = [tool]

    from langchain.agents.agent_toolkits import create_conversational_retrieval_agent
    from langchain.chat_models import ChatOpenAI

    open_api_key = "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"
    llm = ChatOpenAI(temperature=0, openai_api_key=open_api_key)
    agent_executor = create_conversational_retrieval_agent(llm, tools, verbose=True)

    result = agent_executor(
        {"input": "what did the president say about going to east of Columbus?"}
    )
    result["output"]



import traceback

from johnsnowlabs.abstract_base.lib_resolver import try_import_lib
from johnsnowlabs.utils.env_utils import reverse_compatibility_import
from johnsnowlabs.utils.print_messages import log_broken_lib

try:
    if try_import_lib("sparknlp_jsl") and try_import_lib("sparknlp"):
        from sparknlp_jsl.functions import *
        from sparknlp_jsl.training import *
        from sparknlp_jsl.annotator.windowed.windowed_sentence import (
            WindowedSentenceModel,
        )
        from sparknlp_jsl.legal.token_classification.ner.zero_shot_ner import ZeroShotNerModel
        from sparknlp_jsl.training_log_parser import ner_log_parser

        from sparknlp_jsl.base import FeaturesAssembler
        from sparknlp_jsl.legal import (
            LegalDocumentHashCoder as DocumentHashCoder,
            LegalNerQuestionGenerator as NerQuestionGenerator,
            LegalBertForSequenceClassification as BertForSequenceClassification,
            LegalDocumentMLClassifierModel as DocumentMLClassifierModel,
            LegalDocumentMLClassifierApproach as DocumentMLClassifierApproach,
            DocMapperModel,
            DocMapperApproach,
            LegalBertForSequenceClassification as BertForSequenceClassification,
            LegalBertForTokenClassification as BertForTokenClassification,
            LegalNerApproach as NerApproach,
            LegalNerModel as NerModel,
            LegalClassifierDLApproach as ClassifierDLApproach,
            LegalClassifierDLModel as ClassifierDLModel,
            SentenceEntityResolverModel,
            ChunkMapperModel,
            AssertionDLModel,
            RelationExtractionDLModel,
            ZeroShotRelationExtractionModel,
            ChunkMapperApproach,
            SentenceEntityResolverApproach,
            AssertionDLApproach,
            LegalQuestionAnswering as QuestionAnswering,
            LegalTextGenerator as TextGenerator,
            LegalSummarizer as Summarizer,
            LegalFewShotClassifierModel as FewShotClassifierModel,
            LegalFewShotClassifierApproach as FewShotClassifierApproach,
        )

        # These are licensed annos shared across all libs
        from sparknlp_jsl.annotator import (
            GenericSVMClassifierApproach,
            GenericSVMClassifierModel,
            GenericLogRegClassifierApproach,
            GenericClassifierModel,
            SentenceDetector as TextSplitter,
            MedicalDistilBertForSequenceClassification as DistilBertForSequenceClassification,
            AssertionChunkConverter,
            AssertionLogRegModel,
            DeIdentificationModel,
            DocumentLogRegClassifierModel,
            RelationExtractionModel,
            ChunkMergeModel,
            ChunkMapperModel,
            BertSentenceChunkEmbeddings,
            ChunkKeyPhraseExtraction,
            NerDisambiguatorModel,
            EntityChunkEmbeddings,
            TFGraphBuilder,
            ChunkConverter,
            ChunkFilterer,
            NerConverterInternal,
            NerChunker,
            AssertionFilterer,
            AnnotationMerger,
            RENerChunksFilter,
            ChunkSentenceSplitter,
            ChunkMapperFilterer,
            DateNormalizer,
            GenericClassifierModel,
            ReIdentification,
            Replacer,
            ContextualParserModel,
            AssertionLogRegApproach,
            DeIdentification,
            DocumentLogRegClassifierApproach,
            RelationExtractionApproach,
            ChunkMergeApproach,
            NerDisambiguator,
            ContextualParserApproach,
            GenericClassifierApproach,
            Router,
            AssertionChunkConverter,
            NameChunkObfuscatorApproach,
            NameChunkObfuscator,
            # Resolution2Chunk,
            ResolverMerger,
            NerTemplateRenderModel,
            AverageEmbeddings,
            Doc2ChunkInternal,
            Chunk2Token,
            ExtractiveSummarization,
            ChunkFiltererApproach,
            InternalDocumentSplitter as DocumentSplitter,
            Text2SQL,
            IOBTagger,
            DocumentFiltererByClassifier,
        
        )
        from sparknlp_jsl.modelTracer import ModelTracer

        from sparknlp_jsl.structured_deidentification import StructuredDeidentification

        from sparknlp_jsl import training_log_parser, Deid

        from sparknlp_jsl.compatibility import Compatibility
        from sparknlp_jsl.pretrained import InternalResourceDownloader
        from sparknlp_jsl.eval import (
            NerDLMetrics,
            NerDLEvaluation,
            SymSpellEvaluation,
            POSEvaluation,
            NerCrfEvaluation,
            NorvigSpellEvaluation,
        )


except Exception as err:
    if try_import_lib("sparknlp_jsl") and try_import_lib("sparknlp"):
        log_broken_lib("Enterprise Finance")
        print(f"Error Message : {err}")
        print(f"Error Trace: {traceback.format_exc()}")
        print("Performing reverse compatibility import for legal module")
        reverse_compatibility_import(__file__, globals())



import importlib
import site
import subprocess
from importlib import reload

from johnsnowlabs.py_models.lib_version import LibVersion
from johnsnowlabs.utils.venv_utils import VenvWrapper

reload(site)
import os
from typing import Optional

from johnsnowlabs.utils.enums import LatestCompatibleProductVersion, PyInstallTypes
from johnsnowlabs.py_models.primitive import LibVersionIdentifier
from johnsnowlabs.py_models.jsl_secrets import JslSecrets

import json
import sys
from urllib import request
from pkg_resources import parse_version


def get_all_lib_version_on_pypi(pkg_name):
    url = f"https://pypi.python.org/pypi/{pkg_name}/json"
    releases = json.loads(request.urlopen(url).read())["releases"]
    return sorted(releases, key=parse_version, reverse=True)


def get_latest_lib_version_on_pypi(pkg_name):
    return get_all_lib_version_on_pypi(pkg_name)[0]


def get_pip_lib_version(lib: str, py_exec: str = sys.executable):
    # Get lib version of a library according to pip
    r = subprocess.run([py_exec, "-m", "pip", "list"], capture_output=True, text=True)
    matches = list(filter(lambda x: x.split(" ")[0] == lib, r.stdout.split("\n")))
    if not matches:
        return False  # raise ValueError(f'Could not find lib {lib}')
    else:
        return LibVersion(matches[0].split(" ")[-1])


def uninstall_lib(pip_package_name, py_path=sys.executable):
    cmd = f"{py_path} -m pip uninstall {pip_package_name} -y "
    os.system(cmd)
    reload(site)


def install_standard_pypi_lib(
    pypi_name: str,
    module_name: Optional[str] = None,
    python_path: str = sys.executable,
    upgrade: bool = True,
    re_install: bool = False,
    version: Optional[str] = None,
    download_folder: Optional[str] = None,
    include_dependencies: bool = True,
):
    """
    Install module via pypi.
    runs the command :
        `python -m pip install [module_name]`
        `python -m pip install [module_name] --upgrade`
    :param re_install:
    :param version:
    :param pypi_name: file_name of pypi package or path to local whl
    :param module_name: If defined will import module into globals, making it available to running process
    :param python_path: Which Python to use for installing. Defaults to the Python calling this method.
    :param upgrade: use --upgrade flag or not
    :return:
    """
    if isinstance(version, LibVersion):
        version = version.as_str()

    if not pypi_name:
        raise Exception(
            f"Tried to install software which has no pypi file_name! Aborting."
        )
    print(f"Installing {pypi_name} to {python_path}")
    c = f"{python_path} -m pip install {pypi_name}"
    if version:
        c = c + f"=={version} "
    else:
        c = c + " "

    if upgrade:
        c = c + "--upgrade "
    if re_install:
        c = c + " --force-reinstall"

    if download_folder:
        if version:
            c = f"{python_path} -m pip download {pypi_name}=={version} -d {download_folder}"
        else:
            c = f"{python_path} -m pip download {pypi_name} -d {download_folder}"

    if not include_dependencies:
        c = c + " --no-deps"

    os.system(c)
    if module_name and not download_folder:
        try:
            # See if install worked
            # importlib.import_module(module_name)
            reload(site)
            globals()[module_name] = importlib.import_module(module_name)
        except ImportError as err:
            print(f"Failure Installing {pypi_name}")
            return False
    return True


def install_licensed_pypi_lib(
    secrets: JslSecrets,
    pypi_name,
    module_name,
    product: "AbstractSoftwareProduct",
    spark_version: LibVersionIdentifier = LatestCompatibleProductVersion.pyspark.value,
    upgrade=True,
    py_path: str = sys.executable,
    download_folder: Optional[str] = None,
    include_dependencies: bool = True,
):
    """Install Spark-NLP-Healthcare PyPI Package in target python executable
    This just requires the secret of the library.
    """
    get_deps = True
    missmatch = False
    if "spark-nlp-jsl" in pypi_name or "internal_with_finleg" in pypi_name:
        if not secrets.HC_SECRET:
            return False
        module_name = "sparknlp_jsl"
        secret = secrets.HC_SECRET
        # get_deps = True
    elif "ocr" in pypi_name:
        if not secrets.OCR_SECRET:
            return False
        secret = secrets.OCR_SECRET
        module_name = "sparkocr"
        # get_deps = True

    else:
        raise ValueError(f"Invalid install licensed install target ={pypi_name}")

    try:
        url = product.jsl_url_resolver.get_py_urls(
            secret=secret,
            spark_version_to_match=spark_version,
            install_type=PyInstallTypes.wheel,
        ).url
        cmd = f"{py_path} -m pip install {url}"

        # Install lib
        if upgrade:
            cmd = cmd + " --force-reinstall"
        # cmd = f'{sys.executable} -m pip install {pypi_name}=={lib_version} --extra-index-url https://pypi.johnsnowlabs.com/{secret}'

        if download_folder:
            cmd = f"{py_path} -m pip download {pypi_name} -d {download_folder}"

        if not include_dependencies:
            cmd = cmd + " --no-deps"

        print(f'Running "{cmd.replace(secret, "[LIB_SECRET]")}"')
        os.system(cmd)

        # Check if Install succeeded
        if py_path == sys.executable:
            # Check for python executable that is currently running
            reload(site)
            globals()[module_name] = importlib.import_module(module_name)
        else:
            # Check for python executable which is on this machine but not the same as the running one
            return VenvWrapper.is_lib_in_py_exec(py_path, module_name, False)

    except Exception as err:
        print("Failure to install", err)
        return False
    return True



from johnsnowlabs import try_import_lib

if try_import_lib("haystack"):
    from johnsnowlabs.frameworks.embedding_retrieval.haystack_node import (
        JohnSnowLabsHaystackEmbedder,
        JohnSnowLabsHaystackProcessor,
    )

if try_import_lib("langchain"):
    from johnsnowlabs.frameworks.embedding_retrieval.langchain_node import (
        JohnSnowLabsLangChainEmbedder,
        JohnSnowLabsLangChainCharSplitter,
    )



import os.path

from tests.databricks.db_test_utils import *
from tests.databricks.db_test_utils import (
    assert_job_suc,
    run_cluster_test_suite,
    get_or_create_test_cluster,
)


@db_cloud_node_params
def test_create_cluster_with_custom_libs(creds, node_type):
    lic, host, token = creds
    extra_libs = ["farm-haystack==1.21.2", "langchain"]

    cluster_id = nlp.install_to_databricks(
        json_license_path=lic,
        databricks_host=host,
        databricks_token=token,
        driver_node_type_id=node_type,
        node_type_id=node_type,
        cluster_name="Test Custom Lib Cluster",
        clean_cluster=False,
        block_till_cluster_ready=True,
        extra_pip_installs=extra_libs,
    )
    script = """
import haystack
import langchain
print(haystack)
print(langchain)


    """
    # cluster_id = "1005-134624-a6p9ji9u"

    arg1 = "My first arg"
    arg2 = "My second arg"
    assert_job_suc(
        nlp.run_in_databricks(
            script,
            databricks_cluster_id=cluster_id,
            databricks_host=host,
            databricks_token=token,
            run_name="Notebook Test",
            parameters=[arg1, arg2],
        )
    )



import unittest


def replace_string_in_file(file_path, string_to_find, string_to_replace_with):
    # replace_string_in_file and write it back to disk
    # Read the contents of the file
    with open(file_path, "r") as file:
        file_contents = file.read()

    # Replace the target string
    file_contents = file_contents.replace(string_to_find, string_to_replace_with, 1)

    # Write the modified contents back to the file
    with open(file_path, "w") as file:
        file.write(file_contents)


class BackwardCompatTestCase(unittest.TestCase):
    def test_backward_compat(self):
        f = "./johnsnowlabs/medical.py"
        # 1. Inject a "Deprecated" import statement to the medical module and then import it
        replace_string_in_file(
            f,
            string_to_find="from sparknlp_jsl.training import *",
            string_to_replace_with="from sparknlp_jsl.pretrained import InternalBananaDownloader",
        )

        # 2. Globals should be updated properly
        from johnsnowlabs import medical

        self.assertIsNotNone(medical.DistilBertForSequenceClassification)
        self.assertIsNotNone(medical.log_outdated_lib)
        self.assertIsNotNone(medical.AssertionDLModel)

        # 3. Undo the injection
        replace_string_in_file(
            f,
            string_to_replace_with="from sparknlp_jsl.training import *",
            string_to_find="from sparknlp_jsl.pretrained import InternalBananaDownloader",
        )


if __name__ == "__main__":
    unittest.main()



# DB-API Helper funcs

import base64
import json
import os
import time
from pprint import pprint
from typing import Union, List, Optional

import requests

from johnsnowlabs import nlp

ENDPOINT_MAP = {
    # secret endpoints
    "list_scopes": "api/2.0/secrets/scopes/list",
    "create_scope": "api/2.0/secrets/scopes/create",
    "create_scret": "api/2.0/secrets/put",
    "list_secrets": "api/2.0/secrets/list",
    # serve endpoints
    "update_endpoint": "api/2.0/serving-endpoints/{endpoint_name}/config",
    "endpoint_exists": "api/2.0/serving-endpoints/{endpoint_name}",
    "wait_for_endpoint": "api/2.0/serving-endpoints/{endpoint_name}",
    "create_endpoint": "api/2.0/serving-endpoints",
    "update_endpoint_conf": "api/2.0/serving-endpoints/{endpoint_name}/config",
    "delete_endpoint": "api/2.0/serving-endpoints/{endpoint_name}",
    "list_endpoints": "api/2.0/serving-endpoints",
    "get_endpoint": "api/2.0/serving-endpoints/{endpoint_name}",
}


def get_endpoint(name, host, token):
    """Check if an endpoint with the serving_endpoint_name exists"""
    response = requests.get(
        get_endpoint_url(
            ENDPOINT_MAP["list_endpoints"].format(endpoint_name=name), host
        ),
        headers=get_headers(token),
    )
    return response.json()


def delete_endpoint(endpoint_name, host, token):
    """Check if an endpoint with the serving_endpoint_name exists"""
    response = requests.delete(
        get_endpoint_url(
            ENDPOINT_MAP["delete_endpoint"].format(endpoint_name=endpoint_name), host
        ),
        headers=get_headers(token),
    )
    return response.status_code == 200


def list_endpoints(host, token):
    """Check if an endpoint with the serving_endpoint_name exists"""
    response = requests.get(
        get_endpoint_url(ENDPOINT_MAP["list_endpoints"], host),
        headers=get_headers(token),
    )
    return response.json()


def delete_all_endpoints(host, token):
    for end in list_endpoints(host, token)["endpoints"]:
        print("deleting endpoint ", end["name"])
        delete_endpoint(end["name"], host, token)


def dump(data):
    return json.dumps(data).encode("utf-8")


def get_headers(db_token):
    return {"Authorization": f"Bearer {db_token}"}


def get_endpoint_url(req, db_host):
    return f"{db_host}/{req}"


# def update_endpoint(endpoint_name):
#     """Update serving endpoint and wait for it to be ready"""
#     print(f"Updating existing serving endpoint: {endpoint_name}")
#     pay = {"served_models": served_models, "traffic_config": traffic_config}
#     response = requests.post(
#         get_endpoint_url(
#             ENDPOINT_MAP["update_endpoint"].format(endpoint_name=endpoint_name)
#         ),
#         data=dump(pay),
#         headers=get_headers(),
#     )
#
#     response.raise_for_status()
#     wait_for_endpoint()
#     displayHTML(
#         f"""Updated the <a href="/#mlflow/endpoints/{ENDPOINT_NAME}" target="_blank">{ENDPOINT_NAME}</a> serving endpoint"""
#     )


def endpoint_exists(endpoint_name, host, token):
    """Check if an endpoint with the serving_endpoint_name exists"""
    response = requests.get(
        get_endpoint_url(
            ENDPOINT_MAP["endpoint_exists"].format(endpoint_name=endpoint_name), host
        ),
        headers=get_headers(token),
    )
    return response.status_code == 200


def wait_for_endpoint(endpoint_name, host, token):
    """Wait until deployment is ready, then return endpoint config"""
    endpoint_url = get_endpoint_url(
        ENDPOINT_MAP["wait_for_endpoint"].format(endpoint_name=endpoint_name), host
    )
    response = requests.request(
        method="GET", headers=get_headers(token), url=endpoint_url
    )

    print("Deployment starting, this may take 10 to 20 minutes...")
    while (
        response.json()["state"]["ready"] == "NOT_READY"
        or response.json()["state"]["config_update"] == "IN_PROGRESS"
    ):
        if response.json()["state"]["config_update"] == "UPDATE_FAILED":
            print("Something went wrong! Check Serving UI and Status:")
            pprint(response.json())
            raise Exception("Endpoint failed!")
        time.sleep(30)
        response = requests.request(
            method="GET", headers=get_headers(token), url=endpoint_url
        )
        response.raise_for_status()
    return response.json()


def create_endpoint(
    endpoint_name,
    model_name,
    model_version,
    secret_scope_name,
    secret_name,
    db_host,
    token,
    workload_size="Small",
    block_until_deployed=True,
    workload_type="CPU",
):
    """Create serving endpoint and wait for it to be ready
    maps name of your secret to an env variable with the same name in the container
    """
    print(f"Creating new serving endpoint: {endpoint_name}")
    endpoint_url = f"{db_host}/api/2.0/serving-endpoints"
    served_models = [
        {
            "name": "current",
            "model_name": model_name,
            "model_version": model_version,
            "workload_size": workload_size,
            "workload_type": workload_type,
            "scale_to_zero_enabled": "false"
            if "gpu" in workload_type.lower()
            else "true",
            "env_vars": [
                {
                    "env_var_name": "JOHNSNOWLABS_LICENSE_JSON",
                    "secret_scope": secret_scope_name,
                    "secret_key": secret_name,
                },
                {
                    "env_var_name": "spark.databricks.api.url",
                    "secret_scope": secret_scope_name,
                    "secret_key": "DB_API_URL",
                },
                {
                    "env_var_name": "DB_ENDPOINT_ENV",
                    "secret_scope": secret_scope_name,
                    "secret_key": "DB_ENDPOINT_ENV",
                },
            ],
        }
    ]

    request_data = {"name": endpoint_name, "config": {"served_models": served_models}}
    json_bytes = json.dumps(request_data).encode("utf-8")
    response = requests.post(endpoint_url, data=json_bytes, headers=get_headers(token))
    response.raise_for_status()
    if block_until_deployed:
        wait_for_endpoint(endpoint_name, db_host, token)

    try:
        displayHTML(
            f"""Created the <a href="/#mlflow/endpoints/{endpoint_name}" target="_blank">{endpoint_name}</a> serving endpoint"""
        )
    except:
        print(
            f"Created serving endpoint {endpoint_name} at {db_host}/#mlflow/endpoints/{endpoint_name}"
        )


######## Secret handling
def list_scopes(host, db_token):
    response = requests.get(
        get_endpoint_url(ENDPOINT_MAP["list_scopes"], host),
        headers=get_headers(db_token),
    )
    response.raise_for_status()
    return response.text


def create_scope(scope_name, host, db_token):
    pay = {
        "scope": scope_name,
    }
    response = requests.post(
        get_endpoint_url(ENDPOINT_MAP["create_scope"], host),
        data=dump(pay),
        headers=get_headers(db_token),
    )
    print(response.text)
    response.raise_for_status()
    return response.text


def create_secret_in_scope(key, value, scope_name, host, db_token):
    base64_encoded_value = base64.b64encode(value.encode("utf-8")).decode("utf-8")
    pay = {
        "scope": scope_name,
        "key": key,
        "value": value,
        "bytes_value": base64_encoded_value,
    }
    response = requests.post(
        get_endpoint_url(ENDPOINT_MAP["create_scret"], host),
        data=dump(pay),
        headers=get_headers(db_token),
    )
    response.raise_for_status()
    return response.text


def list_secrets_in_scope(scope_name, host, db_token):
    pay = {
        "scope": scope_name,
    }
    response = requests.get(
        get_endpoint_url(ENDPOINT_MAP["list_secrets"], host),
        data=dump(pay),
        headers=get_headers(db_token),
    )
    response.raise_for_status()
    return response.text


def scope_exists(scope_name, host, db_token):
    scopes = json.loads(list_scopes(host, db_token))
    if not scopes:
        return False
    if "scopes" not in scopes:
        return False
    scopes = scopes["scopes"]
    return len(list(filter(lambda x: x["name"] == scope_name, scopes))) == 1


def setup_secrets(scope_name, secret_name, secret_value, host, db_token):
    # 1) create scope
    if not scope_exists(scope_name=scope_name, host=host, db_token=db_token):
        print(f"Scope {scope_name} does not exist, creating it")
        create_scope(scope_name, host, db_token=db_token)
    # 2) Write license
    print(f"Writing license to scope {scope_name}")
    create_secret_in_scope(
        key=secret_name,
        value=secret_value,
        scope_name=scope_name,
        host=host,
        db_token=db_token,
    )
    # 3) Write DB API URL
    create_secret_in_scope(
        key="DB_API_URL",
        value=os.environ["DATABRICKS_HOST"],
        scope_name="JSL_SCOPE",
        host=os.environ["DATABRICKS_HOST"],
        db_token=os.environ["DATABRICKS_TOKEN"],
    )

    # 4) NLU hack
    create_secret_in_scope(
        key="DB_ENDPOINT_ENV",
        value="DB_ENDPOINT_ENV",
        scope_name="JSL_SCOPE",
        host=os.environ["DATABRICKS_HOST"],
        db_token=os.environ["DATABRICKS_TOKEN"],
    )


####### Mlfow Client Utils


def delete_registerd_model(name):
    from mlflow import MlflowClient

    print(f"Deleting registered model {name}")
    return MlflowClient().delete_registered_model(name)


def delete_all_registerd_model(name):
    # TODO
    from mlflow import MlflowClient

    # MlflowClient().list_artifacts()
    return MlflowClient().delete_registered_model(name)


def get_latest_registerd_model_version(name):
    from mlflow import MlflowClient

    try:
        model = MlflowClient().get_registered_model(name)
        return model.latest_versions[-1].version
    except Exception as e:
        print(
            f"Failure getting latest model version! This is expected  on UC environments \n",
            e,
        )
        return "1"


def model_exists(name):
    from mlflow import MlflowClient

    try:
        model = MlflowClient().get_registered_model(name)
    except:
        return False
    return True


############### High level Deployment & Query


def _query_endpoint(data, endpoint_name, db_host, db_token):
    # 5. Query the Endpoint
    # endpoint_name = f"{nlu_model_name.replace('.','_')}_ENDPOINT"

    url = f"{db_host}/serving-endpoints/{endpoint_name}/invocations"
    headers = {
        "Authorization": f"Bearer {db_token}",
        "Content-Type": "application/json",
    }
    response = requests.request(method="POST", headers=headers, url=url, data=data)
    if response.status_code != 200:
        raise Exception(
            f"Request failed with status {response.status_code}, {response.text}"
        )
    import pandas as pd

    return pd.DataFrame(json.loads(response.json()["predictions"]))


def query_to_json(
    in_data: Union[str, List[str]],
    output_level: Optional[str] = None,
    positions: Optional[bool] = None,
    metadata: Optional[bool] = None,
    drop_irrelevant_cols: Optional[bool] = None,
    get_embeddings: Optional[bool] = None,
    keep_stranger_features: Optional[bool] = None,
    multithread: Optional[bool] = None,
):
    data = {}
    data["dataframe_split"] = {}
    data["dataframe_split"]["columns"] = ["text"]
    if output_level:
        data["dataframe_split"]["columns"].append("output_level")
    if positions:
        data["dataframe_split"]["columns"].append("positions")
    if metadata:
        data["dataframe_split"]["columns"].append("metadata")
    if drop_irrelevant_cols:
        data["dataframe_split"]["columns"].append("drop_irrelevant_cols")
    if get_embeddings:
        data["dataframe_split"]["columns"].append("get_embeddings")
    if keep_stranger_features:
        data["dataframe_split"]["columns"].append("keep_stranger_features")
    if multithread:
        data["dataframe_split"]["columns"].append("multithread")

    def expand_data(in_data):
        datas = [in_data]
        if output_level:
            datas.append(output_level)
        if positions:
            datas.append(str(positions))
        if metadata:
            datas.append(str(metadata))
        if drop_irrelevant_cols:
            datas.append(str(drop_irrelevant_cols))
        if get_embeddings:
            datas.append(str(get_embeddings))
        if keep_stranger_features:
            datas.append(str(keep_stranger_features))
        if multithread:
            datas.append(str(multithread))
        return datas

    if isinstance(in_data, str):
        data["dataframe_split"]["data"] = [expand_data(in_data)]
    elif isinstance(in_data, list):
        # data["dataframe_split"]["data"] = [[s,'document'] for s in in_data]
        data["dataframe_split"]["data"] = [expand_data(s) for s in in_data]

    else:
        raise Exception("Input must be str or list of str ")
    return json.dumps(data)


def log_nlu_model(nlu_model_name, registerd_model_name, gpu):
    import mlflow

    # 1. Load the model
    if isinstance(nlu_model_name, str):
        nlu_model = nlp.load(nlu_model_name)
    else:
        # pre-loaded pipe
        nlu_model = nlu_model_name

    # 2. Log the model
    mlflow.johnsnowlabs.log_model(
        nlu_model, "model", registered_model_name=registerd_model_name, gpu=gpu
    )

    # # 3. Download wheels to the model (current version +1)
    try:
        mlflow.models.utils.add_libraries_to_model(
            f"models:/{registerd_model_name}/latest",
            registered_model_name=registerd_model_name,
        )
    except Exception as e:
        # /latest is not supported on UC environments
        mlflow.models.utils.add_libraries_to_model(
            f"models:/{registerd_model_name}/1",
            registered_model_name=registerd_model_name,
        )


def nlu_name_to_endpoint(nlu_model_name):
    return f"{nlu_model_name.replace('.', '_')}_ENDPOINT"


def nlu_name_to_registerd_model(nlu_model_name):
    return f"{nlu_model_name.replace('.', '_')}_REGISTERD_MODEL"


def is_nlu_pipe(pipe):
    from nlu.pipe.pipeline import NLUPipeline

    return isinstance(pipe, NLUPipeline)


def validate_db_creds(db_host, db_token):
    if not db_host:
        db_host = os.environ.get("DATABRICKS_HOST")
    if not db_token:
        db_token = os.environ.get("DATABRICKS_TOKEN")
    if not db_host:
        raise Exception(
            "You must specify DATABRICKS_HOST and DATABRICKS_TOKEN en variables"
        )
    return db_host, db_token


def deploy_model(
    model,
    re_create_endpoint,
    re_create_model,
    db_host,
    db_token,
    workload_size,
    block_until_deployed,
    gpu,
    workload_type,
    endpoint_name,
):
    # convert to nlu pipe if needed
    if isinstance(model, str):
        return deploy_nlu_model_as_endpoint(
            model,
            re_create_endpoint=re_create_endpoint,
            re_create_model=re_create_model,
            db_host=db_host,
            db_token=db_token,
            workload_size=workload_size,
            block_until_deployed=block_until_deployed,
            gpu=gpu,
            workload_type=workload_type,
            endpoint_name=endpoint_name,
        )
    else:
        if not endpoint_name:
            raise Exception(
                "If you want to deploy custom pipes, you need to specify a endpoint_name"
            )
        try:
            import nlu

            if not isinstance(model, nlu.NLUPipeline):
                model = nlp.to_nlu_pipe(model)
        except:
            raise Exception("Failure converting your model to NLU pipe")
        return deploy_nlu_model_as_endpoint(
            model,
            re_create_endpoint=re_create_endpoint,
            re_create_model=re_create_model,
            endpoint_name=endpoint_name,
            db_host=db_host,
            db_token=db_token,
            workload_size=workload_size,
            block_until_deployed=block_until_deployed,
            gpu=gpu,
            workload_type=workload_type,
        )


def query_and_deploy_if_missing(
    model,
    query,
    re_create_endpoint=False,
    re_create_model=False,
    endpoint_name=None,
    is_json_query=False,
    db_host=None,
    db_token=None,
    workload_size="Small",
    new_run=True,
    block_until_deployed=True,
    gpu=False,
    workload_type="CPU",
    # NLU Predict params
    output_level: Optional[str] = None,
    positions: Optional[bool] = None,
    metadata: Optional[bool] = None,
    drop_irrelevant_cols: Optional[bool] = None,
    get_embeddings: Optional[bool] = None,
    keep_stranger_features: Optional[bool] = None,
    multithread: Optional[bool] = None,
):
    """

    Using the NLU predict() https://nlp.johnsnowlabs.com/docs/en/jsl/predict_api
    and to_nlu_pipeline()  https://nlp.johnsnowlabs.com/docs/en/jsl/utils_for_spark_nlp#nlptonlupipepipe


    nlu_model: reference to nlu_model you want to query or  NLU convertable pipe
    Supported types are
    - List[Annotator]
    - Pipeline
    - LightPipeline
    - PretrainedPipeline
    - PipelineModel
    - NLUPipeline
    - String Reference to NLU Pipeline name
        See https://nlp.johnsnowlabs.com/docs/en/jsl/utils_for_spark_nlp#nlptonlupipepipe for more details

    query: str or list of strings or raw json string. If raw json, is_json_query must be True
    is_json_query: if True, query is treated as raw json string
    endpoint_name: Name-Prefix for all resources created (Endpoints, Models, etc). If using non nlu referenced based models, you must specify this.
    re_create_endpoint: if False, endpoint creation is skipped if one already exists. If True, it will delete existing endpoint if it exists
    re_create_model: if False, model creation is skipped if one already exists. If True, model will be re-logged again, bumping the current version by 2
    workload_size: one of Small, Medium, Large.
    new_run: if True, mlflow will start a new run before logging the model
    db_host: the databricks host URL. If not specified, the DATABRICKS_HOST environment variable is used
    db_token: the databricks Access Token. If not specified, the DATABRICKS_TOKEN environment variable is used
    block_until_deployed: if True, this function will block until the endpoint is deployed. If False, it will return immediately after the endpoint is created
    gpu: Use GPU for inference
    workload_type: 'CPU' or  'GPU_SMALL' or see official docs
    output_level : token, chunk, sentence, relation, document
    positions: include or exclude character index position of predictions
    metadata: include additional metadata
    drop_irrelevant_cols: drop irrelevant columns
    get_embeddings: Include embedding or not
    keep_stranger_features: Return columns not named "text", 'image" or "file_type"
    multithread:  Use multi-Threading for inference
    """
    print(
        "query_and_deploy_if_missing is deprecated. It will be dropped in johnsnowlabs==5.2.0 Please use nlp.deploy_endpoint() and nlp.query_endpoint() instead."
    )
    db_host, db_token = validate_db_creds(db_host, db_token)

    if gpu and workload_type == "CPU":
        raise ValueError(
            f'When setting gpu=True you must specify a GPU workload_type. I.e. nlp.query_and_deploy(...,workload_type="GPU_SMALL")'
            f"Check official databricks docs for alternative values "
        )

    if output_level and output_level not in [
        "token",
        "chunk",
        "sentence",
        "relation",
        "document",
    ]:
        raise Exception(
            "output_level must be one of token, chunk, sentence, relation, document"
        )
    if workload_size not in ["Small", "Medium", "Large"]:
        print(
            "WARNING! workload_size should be one of Small, Medium, Large for most users."
        )
    if new_run:
        import mlflow

        mlflow.end_run()
        mlflow.start_run()

    endpoint_name = deploy_model(
        model,
        re_create_endpoint=re_create_endpoint,
        re_create_model=re_create_model,
        db_host=db_host,
        db_token=db_token,
        workload_size=workload_size,
        block_until_deployed=block_until_deployed,
        gpu=gpu,
        workload_type=workload_type,
        endpoint_name=endpoint_name,
    )

    if not block_until_deployed:
        return

    query = (
        query
        if is_json_query
        else query_to_json(
            in_data=query,
            output_level=output_level,
            positions=positions,
            metadata=metadata,
            drop_irrelevant_cols=drop_irrelevant_cols,
            get_embeddings=get_embeddings,
            keep_stranger_features=keep_stranger_features,
            multithread=multithread,
        )
    )
    return query_endpoint(
        endpoint_name=endpoint_name,
        query=query,
        is_json_query=is_json_query,
        output_level=output_level,
        positions=positions,
        metadata=metadata,
        drop_irrelevant_cols=drop_irrelevant_cols,
        get_embeddings=get_embeddings,
        keep_stranger_features=keep_stranger_features,
        multithread=multithread,
        db_host=db_host,
        db_token=db_token,
    )


def deploy_nlu_model_as_endpoint(
    model_name,
    re_create_endpoint=False,
    re_create_model=False,
    endpoint_name=None,
    db_host=None,
    db_token=None,
    workload_size="Small",
    block_until_deployed=True,
    gpu=False,
    workload_type="CPU",
):
    os.environ["MLFLOW_WHEELED_MODEL_PIP_DOWNLOAD_OPTIONS"] = "--prefer-binary"
    SCOPE_NAME = "JSL_SCOPE"
    SECRET_NAME = "JSL_SECRET_NAME"
    SECRET_VALUE = os.environ["JOHNSNOWLABS_LICENSE_JSON_FOR_CONTAINER"]
    REGISTERD_MODEL_NAME = (
        endpoint_name if endpoint_name else nlu_name_to_registerd_model(model_name)
    )
    ENDPOINT_NAME = endpoint_name if endpoint_name else nlu_name_to_endpoint(model_name)

    if not model_exists(REGISTERD_MODEL_NAME) or re_create_model:
        # 1. Log the model
        if model_exists(REGISTERD_MODEL_NAME):
            delete_registerd_model(REGISTERD_MODEL_NAME)
        log_nlu_model(model_name, REGISTERD_MODEL_NAME, gpu=gpu)
    else:
        print(
            "Model already has been logged, skipping logging and using latest. Set re_create_model=True if you want to cre-create it"
        )

    MODEL_VERSION = get_latest_registerd_model_version(REGISTERD_MODEL_NAME)

    if not endpoint_exists(ENDPOINT_NAME, db_host, db_token) or re_create_endpoint:
        if endpoint_exists(ENDPOINT_NAME, db_host, db_token):
            print(f"Deleting exisiting Endpoint {ENDPOINT_NAME}")
            delete_endpoint(ENDPOINT_NAME, db_host, db_token)

        # 2. Create endpboint & Secret Scope if missing. TOdo detect if missing?
        setup_secrets(
            secret_name=SECRET_NAME,
            secret_value=SECRET_VALUE,
            scope_name=SCOPE_NAME,
            host=db_host,
            db_token=db_token,
        )

        # 3. Deploy Endpoint
        create_endpoint(
            endpoint_name=ENDPOINT_NAME,
            model_name=REGISTERD_MODEL_NAME,
            model_version=MODEL_VERSION,
            secret_scope_name=SCOPE_NAME,
            secret_name=SECRET_NAME,
            db_host=db_host,
            token=db_token,
            workload_size=workload_size,
            block_until_deployed=block_until_deployed,
            workload_type=workload_type,
        )
    else:
        print(
            f"Endpoint {ENDPOINT_NAME} already exists!  Set re_create_endpoint=True if you want to re-create it "
        )
    return ENDPOINT_NAME


def deploy_endpoint(
    model,
    re_create_endpoint=False,
    re_create_model=False,
    endpoint_name=None,
    db_host=None,
    db_token=None,
    workload_size="Small",
    new_run=True,
    block_until_deployed=True,
    gpu=False,
    workload_type="CPU",
):
    """
    Using to_nlu_pipeline()  https://nlp.johnsnowlabs.com/docs/en/jsl/utils_for_spark_nlp#nlptonlupipepipe

    nlu_model: reference to nlu_model you want to query or  NLU convertable pipe
    Supported types are
    - List[Annotator]
    - Pipeline
    - LightPipeline
    - PretrainedPipeline
    - PipelineModel
    - NLUPipeline
    - String Reference to NLU Pipeline name
        See https://nlp.johnsnowlabs.com/docs/en/jsl/utils_for_spark_nlp#nlptonlupipepipe for more details

    endpoint_name: Name-Prefix for all resources created (Endpoints, Models, etc). If using non nlu referenced based models, you must specify this.
    re_create_endpoint: if False, endpoint creation is skipped if one already exists. If True, it will delete existing endpoint if it exists
    re_create_model: if False, model creation is skipped if one already exists. If True, model will be re-logged again, bumping the current version by 2
    workload_size: one of Small, Medium, Large.
    new_run: if True, mlflow will start a new run before logging the model
    db_host: the databricks host URL. If not specified, the DATABRICKS_HOST environment variable is used
    db_token: the databricks Access Token. If not specified, the DATABRICKS_TOKEN environment variable is used
    block_until_deployed: if True, this function will block until the endpoint is deployed. If False, it will return immediately after the endpoint is created
    gpu: Use GPU for inference
    workload_type: 'CPU' or  'GPU_SMALL' or see official docs
    """

    db_host, db_token = validate_db_creds(db_host, db_token)
    if workload_size not in ["Small", "Medium", "Large"]:
        print(
            "WARNING! workload_size should be one of Small, Medium, Large for most users."
        )

    if new_run:
        import mlflow

        mlflow.end_run()
        mlflow.start_run()

    return deploy_model(
        model,
        re_create_endpoint=re_create_endpoint,
        re_create_model=re_create_model,
        db_host=db_host,
        db_token=db_token,
        workload_size=workload_size,
        block_until_deployed=block_until_deployed,
        gpu=gpu,
        workload_type=workload_type,
        endpoint_name=endpoint_name,
    )


def query_endpoint(
    endpoint_name,
    query,
    is_json_query=False,
    output_level: Optional[str] = None,
    positions: Optional[bool] = None,
    metadata: Optional[bool] = None,
    drop_irrelevant_cols: Optional[bool] = None,
    get_embeddings: Optional[bool] = None,
    keep_stranger_features: Optional[bool] = None,
    multithread: Optional[bool] = None,
    db_host=None,
    db_token=None,
):
    """
    Using the NLU predict() https://nlp.johnsnowlabs.com/docs/en/jsl/predict_api inside a Databricks Endpoint

    nlu_model: reference to nlu_model you want to query or  NLU convertable pipe
    Supported types are
    - List[Annotator]
    - Pipeline
    - LightPipeline
    - PretrainedPipeline
    - PipelineModel
    - NLUPipeline
    - String Reference to NLU Pipeline name
        See https://nlp.johnsnowlabs.com/docs/en/jsl/utils_for_spark_nlp#nlptonlupipepipe for more details

    query: str or list of strings or raw json string. If raw json, is_json_query must be True
    is_json_query: if True, query is treated as raw json string
    endpoint_name: Name-Prefix for all resources created (Endpoints, Models, etc). If using non nlu referenced based models, you must specify this.
    db_host: the databricks host URL. If not specified, the DATABRICKS_HOST environment variable is used
    db_token: the databricks Access Token. If not specified, the DATABRICKS_TOKEN environment variable is used
    output_level : token, chunk, sentence, relation, document
    positions: include or exclude character index position of predictions
    metadata: include additional metadata
    drop_irrelevant_cols: drop irrelevant columns
    get_embeddings: Include embedding or not
    keep_stranger_features: Return columns not named "text", 'image" or "file_type"
    multithread:  Use multi-Threading for inference"""

    db_host, db_token = validate_db_creds(db_host, db_token)
    query = (
        query
        if is_json_query
        else query_to_json(
            in_data=query,
            output_level=output_level,
            positions=positions,
            metadata=metadata,
            drop_irrelevant_cols=drop_irrelevant_cols,
            get_embeddings=get_embeddings,
            keep_stranger_features=keep_stranger_features,
            multithread=multithread,
        )
    )
    return _query_endpoint(
        data=query,
        endpoint_name=endpoint_name,
        db_host=db_host,
        db_token=db_token,
    )



import json
from multiprocessing import Queue
from threading import Thread

from johnsnowlabs.auto_install.jsl_home import get_install_suite_from_jsl_home
from johnsnowlabs.utils.enums import JvmHardwareTarget
from tests.databricks.db_test_utils import *
from tests.databricks.db_test_utils import (
    run_endpoint_tests,
    get_or_create_test_cluster,
    subtester_thread,
)


def log_and_get_failed_models(results):
    retry_models = []
    for model, result in results.items():
        print(f"Model {model}: {result}")
        if result["success"] is False:
            retry_models.append(model)
    return retry_models


def parallel_run(
    cluster_ids,
    n_parallel_jobs_per_cluster,
    models_to_test,
    host,
    token,
    results,
    test_type,
):
    # 3) For each cluster, start a tester-thread.
    # Start an extra thread for same cluster, for every parallel job run on cluster
    # Threads take jobs from the queue and run them on the cluster till completion.
    job_que = Queue()
    for model in models_to_test:
        job_que.put(model)
    threads = []
    for cluster_id in cluster_ids:
        for i in range(n_parallel_jobs_per_cluster):
            # Start 1 thread for every job that should run, for every cluster
            t = Thread(
                target=subtester_thread,
                args=(
                    cluster_id,
                    job_que,
                    host,
                    token,
                    results,
                    test_type,
                ),
            )
            threads.append(t)
            t.start()
    # Wait for all threads to finish
    for t in threads:
        t.join()


# @pytest.mark.skip(reason="WIP")
@db_cloud_node_params
def test_endpoints_multi_cluster(creds, node_type):
    n_clusters = 1
    n_parallel_jobs_per_cluster = 2
    runtime = "9.1.x-scala2.12"
    lic, host, token = creds

    # 1) Create clusters
    cluster_ids = [
        get_or_create_test_cluster(creds, node_type, i, runtime=runtime)
        for i in range(n_clusters)
    ]

    # 2) Define models to test
    models_to_test = get_mm_models()  # [:3]
    models_to_test = ["tokenize"]
    # one_model_per_class = get_one_model_per_class()

    # 3) Start parallel-job-cluster test
    results = {}
    # test_type = "load_predict"  # 'endpoint'
    test_type = "endpoint"  # ''
    parallel_run(
        cluster_ids=cluster_ids,
        n_parallel_jobs_per_cluster=n_parallel_jobs_per_cluster,
        models_to_test=models_to_test,
        host=host,
        token=token,
        results=results,
        test_type=test_type,
    )

    retry_models = log_and_get_failed_models(results)
    print(f"Retrying {len(retry_models)} models")
    # Give clusters some time to recover from any failures
    time.sleep(60 * 5)

    # run failed models again, with job-parallelism 1 but same cluster-parallelism
    parallel_run(
        cluster_ids=cluster_ids,
        n_parallel_jobs_per_cluster=1,
        models_to_test=retry_models,
        host=host,
        token=token,
        results=results,
        test_type=test_type,
    )
    json.dump(results, open("results.json", "w"))

    # 5) Delete all clusters
    # for cluster_id in cluster_ids:
    #     delete_cluster(cluster_id)


@db_cloud_node_params
def test_endpoint(creds, node_type):
    lic, host, token = creds
    cluster_id = get_or_create_test_cluster(creds, node_type, 10, clean_workspace=False)
    job_url, success = run_endpoint_tests(cluster_id, host, token, "tokenize")
    assert success


@db_cloud_node_params
def test_endpoint_licensed(creds, node_type):
    lic, host, token = creds
    cluster_id = get_or_create_test_cluster(creds, node_type, 0, clean_workspace=True)
    # run_endpoint_tests(cluster_id, host, token, "med_ner.clinical")
    # run_endpoint_tests(cluster_id, host, token, "en.med_ner.clinical")
    job_url, success = run_endpoint_tests(cluster_id, host, token, "tokenize")
    assert success


"""
We want to be able to test on :  
- List of models
- List of runtimes
- auto-generate benchmarks --> time for load/predicting

# TODO
- handle spot instance reclamation https://dbc-3d4c44aa-a512.cloud.databricks.com/?o=4085846932608579#job/407841450144996/run/583312836892114 -162028-g0yy9b85
- handle endpoint stuck/ use timeout 60mins 


"""
# https://github.com/qdrant/qdrant-haystack/tree/master



import os
import sys
from typing import Any, List

from langchain.embeddings.base import Embeddings

# from langchain.pydantic_v1 import BaseModel, Extra
from pydantic import BaseModel, Extra

from johnsnowlabs.frameworks.embedding_retrieval.utils import get_docsplitter_pipe


class JohnSnowLabsLangChainEmbedder(BaseModel, Embeddings):
    """JohnSnowLabs embedding models

    To use, you should have the ``johnsnowlabs`` python package installed.
    Example:        .. code-block:: python
            from langchain.embeddings.johnsnowlabs import JohnSnowLabsEmbeddings
            document = "foo bar"
            embedding = JohnSnowLabsEmbeddings('embed_sentence.bert')
            output = embedding.embed_query(document)
    """

    async def aembed_documents(self, texts: List[str]) -> List[List[float]]:
        raise NotImplementedError("JohnSnowLabsEmbeddings does not support async yet")

    async def aembed_query(self, text: str) -> List[float]:
        raise NotImplementedError("JohnSnowLabsEmbeddings does not support async yet")

    model: Any

    def __init__(
        self, model="embed_sentence.bert", hardware_target="cpu", **kwargs: Any
    ):
        """Initialize the johnsnowlabs model."""
        super().__init__(**kwargs)
        # 1) Check imports
        try:
            from johnsnowlabs import nlp
            from nlu.pipe.pipeline import NLUPipeline
        except ImportError as exc:
            raise ImportError(
                "Could not import johnsnowlabs python package. "
                "Please install it with `pip install johnsnowlabs`."
            ) from exc

            # 2) Start a Spark Session
        try:
            os.environ["PYSPARK_PYTHON"] = sys.executable
            os.environ["PYSPARK_DRIVER_PYTHON"] = sys.executable
            nlp.start(hardware_target=hardware_target)
        except Exception as exc:
            raise Exception("Failure starting Spark Session") from exc
            # 3) Load the model
        try:
            if isinstance(model, str):
                self.model = nlp.load(model)
            elif isinstance(model, NLUPipeline):
                self.model = model
            else:
                self.model = nlp.to_nlu_pipe(model)
        except Exception as exc:
            raise Exception("Failure loading model") from exc

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Compute doc embeddings using a JohnSnowLabs transformer model.

        Args:            texts: The list of texts to embed.
        Returns:            List of embeddings, one for each text."""

        return self.model.predict_embeds(texts)

    def embed_query(self, text: str) -> List[float]:
        """Compute query embeddings using a JohnSnowLabs transformer model.
        Args:            text: The text to embed.
        Returns:            Embeddings for the text."""
        return self.model.predict_embeds(text)[0]


class JohnSnowLabsLangChainCharSplitter:
    def __init__(
        self,
        chunk_overlap=2,
        chunk_size=20,
        explode_splits=True,
        keep_seperators=True,
        patterns_are_regex=False,
        split_patterns=["\n\n", "\n", " ", ""],
        trim_whitespace=True,
    ):
        self.pipe = get_docsplitter_pipe(
            chunk_overlap,
            chunk_size,
            explode_splits,
            keep_seperators,
            patterns_are_regex,
            split_patterns,
            trim_whitespace,
        )

    def split_documents(self, docs):
        return split_lang_docs(self.pipe, docs)


def split_lang_doc(pipe, doc):
    from langchain.schema.document import Document

    return [
        Document(page_content=split, metadata=doc.metadata)
        for split in pipe.annotate(doc.page_content)["splits"]
    ]


def split_lang_docs(pipe, docs):
    return [split for doc in docs for split in split_lang_doc(pipe, doc)]



import os
import sys
import unittest

import tests.utils.secrets as sct
from johnsnowlabs import *
from johnsnowlabs.auto_install.databricks.install_utils import *
from johnsnowlabs.auto_install.jsl_home import get_install_suite_from_jsl_home
from johnsnowlabs.utils.print_messages import log_outdated_lib
from johnsnowlabs.utils.venv_utils import VenvWrapper
from tests.utils import secrets


class AutoInstallTestCase(unittest.TestCase):
    venv_creation_dir = "/home/ckl/old_home/ckl/Documents/freelance/johnsnowlabs_lib/tmp/venv/tmp_test_venv"
    zip_dir = "/home/ckl/Documents/freelance/jsl/johnsnowlabs/tmp/offline"

    def test_quick_bad(self):
        nlp.settings.enforce_versions = False
        nlp.install(enterprise_nlp_secret=sct.enterprise_nlp_sct)

    def test_install_to_databricks_creating_new_cluster(self):
        cluster_id = nlp.install(
            json_license_path=sct.db_lic,
            databricks_host=sct.ckl_host,
            databricks_token=sct.ckl_token,
        )

    def test_install_to_databricks_existing_cluster(self):
        # TODO WIP
        cluster_id = "1006-022913-lb94q2m0"
        nlp.install(
            json_license_path=sct.db_lic,
            databricks_host=sct.ckl_host,
            databricks_cluster_id=cluster_id,
        )
        db = get_db_client_for_token(sct.ckl_host, sct.ckl_token)
        # install_py_lib_via_pip(db, cluster_id, 'nlu')

    def test_install_to_current_env_browser_pop_up(self):
        nlp.install(force_browser=False, visual=True, remote_license_number=2)

    def test_quick(self):
        nlp.install(force_browser=False, visual=True, remote_license_number=2)

    def test_install_pr_secret(self):
        settings.enforce_versions = False
        nlp.install(enterprise_nlp_secret=secrets.pr_secret)

    def test_install_to_current_env(self):
        settings.enforce_versions = False
        nlp.install(json_license_path=sct.old_lic, refresh_install=True)
        # import sparknlp

    def test_install_to_different_python_env(self):
        # Install to env which is not the one we are currently running
        os.system(f"rm -r {self.venv_creation_dir} ")
        f = "/home/ckl/old_home/ckl/Documents/freelance/johnsnowlabs_lib/tmp/licenses/ocr_40.json"
        VenvWrapper.create_venv(self.venv_creation_dir)
        py_path = VenvWrapper.glob_py_exec_from_venv(self.venv_creation_dir)
        nlp.install(json_license_path=f, python_exec_path=py_path)
        self.assertTrue(VenvWrapper.is_lib_in_venv(self.venv_creation_dir, "sparknlp"))
        self.assertTrue(VenvWrapper.is_lib_in_venv(self.venv_creation_dir, "sparkocr"))
        self.assertTrue(
            VenvWrapper.is_lib_in_venv(self.venv_creation_dir, "sparknlp_display")
        )
        self.assertTrue(VenvWrapper.is_lib_in_venv(self.venv_creation_dir, "nlu"))
        self.assertTrue(
            VenvWrapper.is_lib_in_venv(self.venv_creation_dir, "internal_with_finleg")
        )  # ---> sparknlp_jsl
        self.assertTrue(
            VenvWrapper.is_lib_in_venv(self.venv_creation_dir, "jsl_tmp")
        )  # --> johnsnowlabs
        os.system(f"rm -r {self.venv_creation_dir} ")

    def test_create_fresh_venv_and_install_to_it(self):
        # let jsl-lib create a fresh venv for us
        os.system(f"rm -r {self.venv_creation_dir} ")
        f = "/home/ckl/old_home/ckl/Documents/freelance/johnsnowlabs_lib/tmp/licenses/ocr_40.json"
        nlp.install(json_license_path=f, venv_creation_path=self.venv_creation_dir)
        self.assertTrue(VenvWrapper.is_lib_in_venv(self.venv_creation_dir, "sparknlp"))
        self.assertTrue(VenvWrapper.is_lib_in_venv(self.venv_creation_dir, "sparkocr"))
        self.assertTrue(
            VenvWrapper.is_lib_in_venv(self.venv_creation_dir, "sparknlp_display")
        )
        self.assertTrue(VenvWrapper.is_lib_in_venv(self.venv_creation_dir, "nlu"))
        self.assertTrue(
            VenvWrapper.is_lib_in_venv(self.venv_creation_dir, "sparknlp_jsl")
        )
        self.assertTrue(
            VenvWrapper.is_lib_in_venv(self.venv_creation_dir, "johnsnowlabs")
        )
        os.system(f"rm -r {self.venv_creation_dir} ")

    def test_list_license_status(self):
        nlp.check_health(check_licenses=True)
        nlp.list_remote_licenses()
        nlp.list_local_licenses()

    def test_oudated_message(self):
        from johnsnowlabs.auto_install.softwares import Software

        log_outdated_lib(Software.spark_ocr, "79")

    def test_offline_install_print(self):
        nlp.install(offline=True)

    def test_offline_install_zip(self):
        os.system(f"rm -r {self.zip_dir} ")
        nlp.install(
            offline=True,
            offline_zip_dir=self.zip_dir,
            install_optional=True,
            include_dependencies=True,
        )

    def test_browser_install(self):
        nlp.install(force_browser=True, visual=True, local_license_number=2)
        # nlp.install(local_license_number=2, force_browser=True)

    def test_upgrade_licensed_lib_via_secret_only(self):
        new_secret = secrets.random_secret
        from johnsnowlabs import settings

        settings.enforce_versions = False
        nlp.install(enterprise_nlp_secret=new_secret)

    def test_json_license_install(self):
        nlp.install(json_license_path=sct.latest_lic)

        # import sparkocr

        # nlp.install(json_license_path=old_lic)

    def test_json_license_install_outdated(self):
        nlp.settings.enforce_versions = False
        nlp.install(json_license_path=sct.old_lic)

        # import sparkocr

        # nlp.install(json_license_path=old_lic)

    def test_create_and_install_cluster(self):
        install_suite = get_install_suite_from_jsl_home()
        print(install_suite)

    def test_uninstall_all(self):
        # os.system(old_lic'{sys.py_executable} -py_executable pip uninstall spark-nlp -y')
        # os.system(old_lic'{sys.py_executable} -py_executable pip uninstall spark-nlp-display -y')
        # os.system(old_lic'{sys.py_executable} -py_executable pip uninstall nlu -y')

        # os.system(f"{sys.executable} -m pip uninstall spark-nlp-jsl -y")
        # os.system(f"{sys.executable} -m pip uninstall spark-ocr -y")
        # os.system(old_lic'{sys.py_executable} -py_executable pip uninstall jsl_tmp -y')
        os.system(f"{sys.executable} -m pip uninstall spark-nlp-internal -y")

    def test_install_to_emr(self):
        # Make sure correct aws credentials are configured
        nlp.install_to_emr(
            "us-east-1",
            bootstrap_bucket="ksh-emr-bucket",
            subnet_id="subnet-28754965",
            s3_logs_path="s3://ksh-emr-bucket/logs",
        )

    @classmethod
    def tearDownClass(cls):
        1
        # print("TEARING DOWN")
        # os.system(old_lic'rm -r {cls.venv_creation_dir} ')
        # # os.system(old_lic'rm -r {cls.zip_dir} ')
        #
        # os.system(old_lic'{sys.py_executable} -py_executable pip uninstall spark-nlp-jsl -y')
        # os.system(old_lic'{sys.py_executable} -py_executable pip uninstall spark-nlp-jsl -y')
        # os.system(old_lic'{sys.py_executable} -py_executable pip uninstall spark-ocr -y')
        # os.system(old_lic'{sys.py_executable} -py_executable pip uninstall jsl_tmp -y')
        # os.system(old_lic'{sys.py_executable} -py_executable pip uninstall internal_with_finleg -y')
        #
        # os.system(old_lic'{sys.py_executable} -py_executable pip uninstall spark-nlp -y')
        # os.system(old_lic'{sys.py_executable} -py_executable p ip uninstall spark-nlp-display -y')
        # os.system(old_lic'{sys.py_executable} -py_executable pip uninstall nlu -y')
        # os.system(old_lic'{sys.py_executable} -py_executable pip uninstall pyspark -y')

    def test_refresh_credentials(self):
        # Use this to upgrade all secrets on every license file, if greater
        nlp.install(json_license_path=sct.latest_lic, only_refresh_credentials=True)

    def test_refresh_install(self):
        # Use this to upgrade all secrets on every license file, if greater
        nlp.install(json_license_path=sct.latest_lic, refresh_install=True)


if __name__ == "__main__":
    unittest.main()



# Convert pdf to image
from johnsnowlabs import nlp


def run_test():
    # nlp.start()
    data = ["hello world", "I love apples"]
    model = nlp.load("<MODEL>")
    print("model loaded", model)

    df = model.predict(data)
    for c in df.columns:
        print(df[c])


if __name__ == "__main__":
    run_test()



import queue
import time

import pytest
from nlu.utils.environment.offline_load_utils import NLP_ref_to_NLU_ref

from johnsnowlabs import nlp, settings
from johnsnowlabs.auto_install.databricks.install_utils import (
    cluster_exist_with_name_and_runtime,
    get_db_client_for_token,
    get_cluster_id,
    wait_till_cluster_running,
    _get_cluster_id,
)
from johnsnowlabs.auto_install.health_checks.generate_test import (
    generate_endpoint_test,
)
from johnsnowlabs.auto_install.health_checks.generate_test import (
    generate_load_predict_test,
)
from johnsnowlabs.utils.enums import DatabricksClusterStates
from tests.utilsz import secrets as sct

db_cloud_node_params = pytest.mark.parametrize(
    "creds, node_type",
    [
        ("aws_creds", "aws_cpu_node"),
        # ("azure_creds", "azure_cpu_node"),
        # ("azure_creds", "azure_gpu_node_type"),
    ],
    indirect=["creds", "node_type"],
)
db_cloud_params = pytest.mark.parametrize(
    "creds",
    [
        "aws_creds",
        # "azure_creds",
    ],
    indirect=["creds"],
)


@pytest.fixture()
def azure_gpu_node_type():
    return "Standard_NC4as_T4_v3"


@pytest.fixture()
def azure_cpu_node():
    return "Standard_DS3_v2"


@pytest.fixture()
def aws_cpu_node():
    return "i3.xlarge"


@pytest.fixture()
def azure_creds():
    import tests.utilsz.secrets as sct

    lic = sct.db_lic_azure
    host = sct.ckl_host_azure
    token = sct.ckl_token_azure
    return lic, host, token


@pytest.fixture()
def azure_trial_creds():
    import tests.utilsz.secrets as sct

    lic = sct.azure_trail_lic2
    host = sct.azure_trail_host2
    token = sct.azure_trail_token2
    return lic, host, token


@pytest.fixture()
def aws_creds():
    import tests.utilsz.secrets as sct

    lic = sct.ckl_lic_aws
    host = sct.ckl_host_aws
    token = sct.ckl_token_aws
    return lic, host, token


@pytest.fixture
def creds(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def node_type(request):
    return request.getfixturevalue(request.param)


def host_creds(host):
    if host == "aws":
        return aws_creds()
    elif host == "azure":
        return azure_creds()
    else:
        raise Exception("Unknown host")


def assert_job_suc(state):
    assert state["state"]["result_state"] == "SUCCESS"


def run_endpoint_tests(test_cluster_id, host, token, model):
    # generate job.py script and submit it
    job_id, job_url = nlp.run_in_databricks(
        generate_endpoint_test(model, sct.container_lic_json),
        databricks_cluster_id=test_cluster_id,
        databricks_host=host,
        databricks_token=token,
        run_name=f"endpoint_creation_test_run_{model}",
        return_job_url=True,
    )
    try:
        assert_job_suc(job_id)
    except Exception as e:
        return job_url, False

    return job_url, True


def run_load_predict_tests(test_cluster_id, host, token, model):
    # generate job.py script and submit it
    job_id, job_url = nlp.run_in_databricks(
        generate_load_predict_test(model),
        databricks_cluster_id=test_cluster_id,
        databricks_host=host,
        databricks_token=token,
        run_name=f"load_predict_test_{model}",
        return_job_url=True,
    )

    try:
        assert_job_suc(job_id)
    except Exception as e:
        return job_url, False

    return job_url, True


def run_cluster_test_suite(test_cluster_id, host, token):
    # run test suite again a existing cluster

    # Test modules
    import johnsnowlabs.auto_install.health_checks.hc_test as hc_test
    import johnsnowlabs.auto_install.health_checks.ocr_test as ocr_test
    import johnsnowlabs.auto_install.health_checks.nlp_test as nlp_test

    assert_job_suc(
        nlp.run_in_databricks(
            nlp_test,
            databricks_cluster_id=test_cluster_id,
            databricks_host=host,
            databricks_token=token,
            run_name="nlp_test",
        )
    )

    assert_job_suc(
        nlp.run_in_databricks(
            hc_test,
            databricks_cluster_id=test_cluster_id,
            databricks_host=host,
            databricks_token=token,
            run_name="hc_test",
        )
    )

    assert_job_suc(
        nlp.run_in_databricks(
            ocr_test,
            databricks_cluster_id=test_cluster_id,
            databricks_host=host,
            databricks_token=token,
            run_name="ocr_test",
        )
    )


def get_one_model_per_class():
    # todo actually generate from spellbook
    return [
        # "tokenize",
        "sentiment",
        "ner",
    ]
    #     "bert",
    #     "elmo",
    #     "albert",
    #     "roberta",
    # ]


def delete_all_test_clusters(db_client):
    for i in range(10):
        while _get_cluster_id(db_client, f"TEST_CLUSTER_{i}"):
            cluster_id = _get_cluster_id(db_client, f"TEST_CLUSTER_{i}")
            db_client.cluster.permanent_delete_cluster(cluster_id)
            print(f"Deleted cluster {cluster_id}")
            time.sleep(5)


# TODO add UC cluster!?
def get_or_create_test_cluster(
    creds, node_type, n=0, runtime=None, clean_workspace=False
):
    """
    Creates a cluster with name TEST_CLUSTER_{n} and runtime runtime if it doesn't exist.
    If it exists, it checks if it's running and if not, it starts it.
    "exists" means another cluster with same name and runtime exists.
    """
    # m5d.2xlarge  #
    lic, host, token = creds
    # runtime = "9.1.x-scala2.12"
    if runtime is None:
        runtime = settings.db_spark_version
    cluster_name = f"TEST_CLUSTER_{n}"
    db_client = get_db_client_for_token(host, token)

    # 1) Check if cluster with name and correct runtime exists and it's running state If not, create it.
    if cluster_exist_with_name_and_runtime(db_client, cluster_name, runtime):
        cluster_id = get_cluster_id(db_client, cluster_name, runtime)

        status = DatabricksClusterStates(
            db_client.cluster.get_cluster(
                cluster_id,
            )["state"]
        )

        # if status is terminated, try restart it
        if status == DatabricksClusterStates.TERMINATED:
            db_client.cluster.start_cluster(cluster_id=cluster_id)
            wait_till_cluster_running(db_client, cluster_id)
            status = DatabricksClusterStates(
                db_client.cluster.get_cluster(
                    cluster_id,
                )["state"]
            )
        if status == DatabricksClusterStates.RUNNING:
            return cluster_id
        else:
            print(
                f"Could not find or start {cluster_name} with runtime {runtime} creating a new one"
            )

    else:
        #  no cluster exists, create new one
        cluster_id = nlp.install_to_databricks(
            spark_version=runtime,
            json_license_path=lic,
            databricks_host=host,
            databricks_token=token,
            driver_node_type_id=node_type,
            node_type_id=node_type,
            cluster_name=cluster_name,
            clean_cluster=clean_workspace,
            block_till_cluster_ready=True,
        )
        return cluster_id


test_funcs = {
    "endpoint": run_endpoint_tests,
    "load_predict": run_load_predict_tests,
    # "cluster_test_suite": run_cluster_test_suite,
}


def execute_test(test_type, cluster_id, host, token, model):
    if test_type in test_funcs:
        return test_funcs[test_type](
            cluster_id, host, token, model, raise_on_fail=False
        )
    else:
        raise ValueError(f"Unknown test type {test_type}")


def subtester_thread(cluster_id, job_que, host, token, results, test_type):
    while not job_que.empty():
        try:
            model = job_que.get_nowait()
            result = {"model": model}
        except queue.Empty:
            print("Que Empty Thread done!")
            return
        print(f"Test {model} with {test_type}")
        result["job_url"], result["success"] = test_funcs[test_type](
            cluster_id, host, token, model
        )
        results[model] = result
        print(f"✅ {model} TEST RESULTS {results[model]} ")


mm_models = [
    "ner_chemd_clinical_pipeline",
    "summarizer_clinical_questions_pipeline",
    "summarizer_clinical_guidelines_large_pipeline",
    "summarizer_clinical_jsl_augmented_pipeline",
    "summarizer_clinical_laymen_onnx",
    "summarizer_clinical_laymen",
    "ner_profiling_biobert",
    " ner_jsl_pipeline assertion_jsl_augmented",
    "ner_jsl_langtest_pipeline",
    "ner_living_species_pipeline",
    "ner_jsl_pipeline",
    "sbiobertresolve_cpt_augmented",
    "sbiobertresolve_HPO",
    "sbiobertresolve_icd10cm_augmented_billable_hcc",
    "sbiobertresolve_loinc_augmented",
    "sbiobertresolve_rxnorm_augmented",
    "medication_resolver_transform_pipeline",
    "sbiobertresolve_snomed_findings",
    "umls_disease_syndrome_resolver_pipeline",
    "umls_drug_resolver_pipeline",
    "umls_drug_substance_resolver_pipeline",
    "umls_major_concepts_resolver_pipeline",
    "umls_clinical_findings_resolver_pipeline",
    "bert_sequence_classifier_gender_biobert",
    "ner_cellular_pipeline",
    "summarizer_generic_jsl_pipeline",
    "biogpt_chat_jsl",
    "medical_qa_biogpt",
    "re_test_result_date_pipeline",
    "re_temporal_events_clinical_pipeline",
    "ner_pathogen_pipeline",
    "ner_supplement_clinical_pipeline",
    "clinical_notes_qa_base_onnx",
    "clinical_notes_qa_large_onnx",
    "ner_nihss redl_nihss_biobert",
    "multiclassifierdl_hoc",
    "oncology_general_pipeline",
    "ner_profiling_oncology",
    "ner_oncology_response_to_treatment",
    "ner_oncology_therapy",
    "ner_oncology_test",
    "ner_oncology_tnm",
    "ner_oncology_diagnosis",
    "ner_oncology_anatomy_granular",
    "ner_oncology_posology_langtest_pipeline",
    "ner_radiology_wip_clinical assertion_dl_radiology",
    "ner_chexpert_pipeline",
    "summarizer_radiology_pipeline",
    "ner_risk_factors_biobert_pipeline",
    "ner_sdoh_langtest_pipeline assertion_sdoh_wip",
    "ner_profiling_sdoh assertion_sdoh_wip",
    "ner_sdoh_access_to_healthcare",
    "ner_sdoh_community_condition",
    "ner_sdoh_demographics",
    "ner_sdoh_health_behaviours_problems",
    "ner_sdoh_income_social_status",
    "ner_sdoh_social_environment",
    "ner_sdoh_substance_usage",
    "genericclassifier_sdoh_economics_binary_sbiobert_cased_mli",
    "genericclassifier_sdoh_alcohol_usage_sbiobert_cased_mli",
    "ner_nihss_pipeline",
    "ner_vop assertion_vop_clinical_large",
    "bert_sequence_classifier_vop_hcp_consult",
    "bert_sequence_classifier_vop_self_report",
    "bert_sequence_classifier_vop_sound_medical",
    "bert_sequence_classifier_vop_side_effect",
    "bert_sequence_classifier_vop_drug_side_effect",
    "ner_profiling_vop",
    "zero_shot_ner_roberta",
    "re_zeroshot_biobert",
    "distilbert_base_sequence_classifier_imdb",
    "bert_sequence_classifier_emotion",
    "bert_sequence_classifier_age_news",
    "bert_sequence_classifier_trec_coarse",
    "distilbert_base_sequence_classifier_food",
    "bert_sequence_classifier_multilingual_sentiment",
    "distilbert_multilingual_sequence_classifier_allocine",
    "distilbert_base_sequence_classifier_toxicity",
    "bert_sequence_classifier_toxicity",
    "deberta_v3_xsmall_token_classifier_ontonotes",
    "deberta_v3_small_token_classifier_ontonotes",
    "deberta_v3_base_token_classifier_ontonotes",
    "deberta_v3_large_token_classifier_ontonotes",
    "bert_base_token_classifier_ontonote",
    "bert_large_token_classifier_ontonote",
    "bert_token_classifier_aner",
    "bert_token_classifier_named_entity_recognition_nerkor_hu_hungarian",
    "xlmroberta_ner_uk_ner",
    "sentence_detector_dl",
    "bert_token_classifier_base_chinese_ner",
    "bert_token_classifier_base_turkish_cased_ner",
    "roberta_qa_deepset_base_squad2",
    "bert_qa_arap",
    "camembert_base_qa_fquad",
    "distilbert_qa_base_cased_squadv2",
    "xlm_roberta_qa_xlm_roberta_qa_chaii",
    "bart_large_zero_shot_classifier_mnli",
    "roberta_base_zero_shot_classifier_nli",
    "bert_base_cased_zero_shot_classifier_xnli",
    "distilbert_base_zero_shot_classifier_uncased_mnli",
    "distilbert_base_zero_shot_classifier_turkish_cased_snli",
    "xlm_roberta_large_zero_shot_classifier_xnli_anli",
    "opus_mt_en_zh",
    "opus_mt_en_es",
    "opus_mt_en_fr",
    "opus_mt_en_it",
    "opus_mt_es_en",
    "opus_mt_fr_en",
    "opus_mt_it_en",
    "distilbart_xsum_12_6",
    "distilbart_xsum_6_6",
    "distilbart_cnn_6_6",
    "t5_small",
    "e5_small_v2",
    "e5_base_v2",
    "e5_large_v2",
    "e5_small_v2_quantized",
    "e5_base_v2_quantized",
    "e5_large_v2_quantized",
    "instructor_base",
    "instructor_large",
    "multi_qa_mpnet_base_cos_v1",
    "all_mpnet_base_v2",
]


def get_mm_models():
    # TODO use actual langs and make sure all models are resolved
    nlu_refs = []
    miss = []
    for m in mm_models:
        lang = "en"
        models = m.split(" ")
        nlu_ref = ""
        for model in models:
            partial_ref = NLP_ref_to_NLU_ref(model, lang)
            if partial_ref:
                nlu_ref = nlu_ref + " " + partial_ref
            else:
                print(f"{m} missing!")
                miss.append(m)
                continue
        if nlu_ref:
            nlu_refs.append(nlu_ref)

    print(f"Num Missing : {len(miss)} Num Found : {len(nlu_refs)}")
    print("Missing:", miss)
    print("Found:", nlu_refs)
    return nlu_refs



# Embedder Node compatible with haystack framework
import os
import sys
from pathlib import Path
from typing import List, Optional, Union, Dict, Any, Tuple

import numpy as np
from haystack import BaseComponent
from haystack.document_stores import BaseDocumentStore
from haystack.nodes.retriever import EmbeddingRetriever
from haystack.nodes.retriever._base_embedding_encoder import _BaseEmbeddingEncoder
from haystack.schema import Document, MultiLabel
from tqdm.auto import tqdm

from johnsnowlabs.frameworks.embedding_retrieval.utils import get_docsplitter_pipe


class _JohnsnowlabsEmbeddingEncoder(_BaseEmbeddingEncoder):
    def __init__(self, retriever: "EmbeddingRetriever"):
        # 1) Check imports
        try:
            from johnsnowlabs import nlp
            from nlu.pipe.pipeline import NLUPipeline
        except ImportError as exc:
            raise ImportError(
                "Could not import johnsnowlabs python package. "
                "Please install it with `pip install johnsnowlabs`."
            ) from exc

        # 2) Start a Spark Session
        try:
            os.environ["PYSPARK_PYTHON"] = sys.executable
            os.environ["PYSPARK_DRIVER_PYTHON"] = sys.executable
            nlp.start(hardware_target="gpu" if retriever.use_gpu else "cpu")
        except Exception as exc:
            raise Exception("Failure starting Spark Session") from exc
            # 3) Load the model
        try:
            self.embedding_model = nlp.load(retriever.embedding_model)
        except Exception as exc:
            raise Exception("Failure loading model") from exc

    def embed(self, texts: Union[List[str], str]) -> np.ndarray:
        return np.asarray(
            self.embedding_model.predict_embeds(texts),
            dtype=float,
        )

    def embed_queries(self, queries: List[str]) -> np.ndarray:
        return self.embed(queries)

    def embed_documents(self, docs: List[Document]) -> np.ndarray:
        return self.embed([d.content for d in docs])

    def train(
        **kwargs,
    ):
        raise NotImplementedError("Training not supported")

    def save(self, save_dir: Union[Path, str]):
        raise NotImplementedError("Saving not supported")


class JohnSnowLabsHaystackEmbedder(EmbeddingRetriever):
    def __init__(
        self,
        embedding_model: str,
        use_gpu: bool = False,
        document_store: Optional[BaseDocumentStore] = None,
        **kwargs,
    ):
        inject()
        kwargs["model_format"] = "johnsnowlabs"
        super().__init__(
            embedding_model=embedding_model,
            document_store=document_store,
            use_gpu=use_gpu,
            **kwargs,
        )


class JohnSnowLabsHaystackProcessor(BaseComponent):  # BasePreProcessor
    outgoing_edges = 1

    def __init__(
        self,
        chunk_overlap=2,
        chunk_size=20,
        explode_splits=True,
        keep_seperators=True,
        patterns_are_regex=False,
        split_patterns=["\n\n", "\n", " ", ""],
        trim_whitespace=True,
        ##### OLD PARAMS #####
        progress_bar: bool = True,
        add_page_number: bool = False,
        max_chars_check: int = 10_000,
    ):
        """
        :param clean_header_footer: Use heuristic to remove footers and headers across different pages by searching
                                     for the longest common string. This heuristic uses exact matches and therefore
                                     works well for footers like "Copyright 2019 by XXX", but won't detect "Page 3 of 4"
                                     or similar.
        :param clean_whitespace: Strip whitespaces before or after each line in the text.
        :param clean_empty_lines: Remove more than two empty lines in the text.
        :param remove_substrings: Remove specified substrings from the text. If no value is provided an empty list is created by default.
        :param split_by: Unit for splitting the document. Can be "word", "sentence", or "passage". Set to None to disable splitting.
        :param split_length: Max. number of the above split unit (e.g. words) that are allowed in one document. For instance, if n -> 10 & split_by ->
                           "sentence", then each output document will have 10 sentences.
        :param split_overlap: Word overlap between two adjacent documents after a split.
                              Setting this to a positive number essentially enables the sliding window approach.
                              For example, if split_by -> `word`,
                              split_length -> 5 & split_overlap -> 2, then the splits would be like:
                              [w1 w2 w3 w4 w5, w4 w5 w6 w7 w8, w7 w8 w10 w11 w12].
                              Set the value to 0 to ensure there is no overlap among the documents after splitting.
        :param split_respect_sentence_boundary: Whether to split in partial sentences if split_by -> `word`. If set
                                                to True, the individual split will always have complete sentences &
                                                the number of words will be <= split_length.
        :param language: The language used by "nltk.tokenize.sent_tokenize" in iso639 format.
            Available options: "ru","sl","es","sv","tr","cs","da","nl","en","et","fi","fr","de","el","it","no","pl","pt","ml"
        :param tokenizer_model_folder: Path to the folder containing the NTLK PunktSentenceTokenizer models, if loading a model from a local path. Leave empty otherwise.
        :param id_hash_keys: Generate the document id from a custom list of strings that refer to the document's
            attributes. If you want to ensure you don't have duplicate documents in your DocumentStore but texts are
            not unique, you can modify the metadata and pass e.g. `"meta"` to this field (e.g. [`"content"`, `"meta"`]).
            In this case the id will be generated by using the content and the defined metadata.
        :param progress_bar: Whether to show a progress bar.
        :param add_page_number: Add the number of the page a paragraph occurs in to the Document's meta
                                field `"page"`. Page boundaries are determined by `"\f"` character which is added
                                in between pages by `PDFToTextConverter`, `TikaConverter`, `ParsrConverter` and
                                `AzureConverter`.
        :param max_chars_check: the maximum length a document is expected to have. Each document that is longer than max_chars_check in characters after pre-processing will raise a warning.
        """
        super().__init__()
        self.progress_bar = progress_bar
        self.pipe = get_docsplitter_pipe(
            chunk_overlap,
            chunk_size,
            explode_splits,
            keep_seperators,
            patterns_are_regex,
            split_patterns,
            trim_whitespace,
        )

    def process(
        self,
        documents: Union[dict, Document, List[Union[dict, Document]]] = None,
    ) -> List[Document]:
        """
        Perform document cleaning and splitting. Can take a single document or a list of documents as input and returns a list of documents.
        """
        if isinstance(documents, (Document, dict)):
            ret = self._process_single(document=documents)
        elif isinstance(documents, list):
            ret = self._process_batch(documents=list(documents))
        elif documents is None:
            ret = []
        else:
            raise Exception(
                f"documents provided to PreProcessor.prepreprocess() is not of type list nor Document or None but is {type(documents)} "
            )

        return ret

    def _process_single(
        self,
        document: Union[dict, Document],
    ) -> List[Document]:
        return self.split(document=self.clean(document=document))

    def _process_batch(self, documents: List[Union[dict, Document]]) -> List[Document]:
        nested_docs = [
            self._process_single(d)
            for d in tqdm(
                documents,
                disable=not self.progress_bar,
                desc="Preprocessing",
                unit="docs",
            )
        ]
        return [d for x in nested_docs for d in x]

    def clean(
        self,
        document: Union[dict, Document],
        # not implemented
        clean_whitespace: bool = None,
        clean_header_footer: bool = None,
        clean_empty_lines: bool = None,
        remove_substrings: Optional[List[str]] = None,
    ) -> Document:
        """
        Stub for Feature cleaning procedure
        """
        return document

    def split(
        self,
        document: Union[dict, Document],
    ) -> List[Document]:
        """Perform document splitting on a single document. This method can split on different units, at different lengths,
        with different strides. It can also respect sentence boundaries. Its exact functionality is defined by
        the parameters passed into PreProcessor.__init__(). Takes a single document as input and returns a list of documents.
        """
        # Todo pass params dynamically?
        text = document.content
        texts = [split for split in self.pipe.annotate(text)["splits"]]
        return [
            Document(id_hash_keys=["content", "meta"], content=text, meta=document.meta)
            for i, text in enumerate(texts)
        ]

    def run(
        self,
        query: Optional[str] = None,
        file_paths: Optional[List[str]] = None,
        labels: Optional[MultiLabel] = None,
        documents: Optional[List[Document]] = None,
        meta: Optional[dict] = None,
    ) -> Tuple[Dict, str]:
        """
        Method that will be executed when the node in the graph is called.

        The argument that are passed can vary between different types of nodes
        (e.g. retriever nodes expect different args than a reader node)


        See an example for an implementation in haystack/reader/base/BaseReader.py
        :return:
        """
        # result = {"documents": d.to_dict() for d in self.process(documents)}
        result = {"documents": d for d in self.process(documents)}
        return result, "output_1"

    def run_batch(
        self,
        queries: Optional[Union[str, List[str]]] = None,
        file_paths: Optional[List[str]] = None,
        labels: Optional[Union[MultiLabel, List[MultiLabel]]] = None,
        documents: Optional[Union[List[Document], List[List[Document]]]] = None,
        meta: Optional[Union[Dict[str, Any], List[Dict[str, Any]]]] = None,
        params: Optional[dict] = None,
        debug: Optional[bool] = None,
    ):
        result = {"documents": d for d in self.process(documents)}
        return result, "output_1"


def inject():
    # inject the emd encoder into haystack
    from haystack.nodes.retriever import _embedding_encoder

    _embedding_encoder._EMBEDDING_ENCODERS[
        "johnsnowlabs"
    ] = _JohnsnowlabsEmbeddingEncoder
    # inject the retriever into haystack


inject()

# if id_hash_keys is None:
#     id_hash_keys = self.id_hash_keys
#
# if isinstance(document, dict):
#     document["id_hash_keys"] = id_hash_keys
#     document = Document.from_dict(document)
#
# # Mainly needed for type checking
# if not isinstance(document, Document):
#     raise HaystackError(
#         "Document must not be of type 'dict' but of type 'Document'."
#     )
#
# if type(document.content) is not str:
#     return document
#
# text = document.content
# # TODO APPLY PIPE
# self.pipe.annotate(text)
# texts = [split for split in self.pipe.annotate(text)["splits"]]
# if text != document.content:
#     document = deepcopy(document)
#     document.content = text
# return document



import os
from johnsnowlabs.auto_install.databricks.install_utils import *
# fmt: off
from tests.databricks.db_test_utils import aws_creds
# fmt: on


def test_hdfs_basic_methods(aws_creds):
    lic, host, token = aws_creds
    db_client = get_db_client_for_token(host, token)
    src_p = os.path.abspath(__file__)
    target_p = "/johnsnowlabs/testf"
    copy_from_local_to_hdfs(db_client, local_path=src_p, dbfs_path=target_p)
    copy_from_local_to_hdfs(db_client, local_path=src_p, dbfs_path=target_p + "2")
    copy_from_local_to_hdfs(db_client, local_path=src_p, dbfs_path=target_p + "3")
    pprint(dbfs_ls(db_client, "/johnsnowlabs"))
    dbfs_rm(db_client, target_p)
    pprint(dbfs_ls(db_client, "/johnsnowlabs"))



import traceback

from johnsnowlabs.abstract_base.lib_resolver import try_import_lib
from johnsnowlabs.auto_install.softwares import Software
from johnsnowlabs.utils.env_utils import reverse_compatibility_import
from johnsnowlabs.utils.print_messages import log_outdated_lib, log_broken_lib


warning_logged = False

try:
    if try_import_lib("sparknlp_jsl") and try_import_lib("sparknlp"):
        from sparknlp_jsl.functions import *
        from sparknlp_jsl.training import *
        from sparknlp_jsl.annotator.windowed.windowed_sentence import (
            WindowedSentenceModel,
        )
        from sparknlp_jsl.annotator.ner.zero_shot_ner import ZeroShotNerModel
        from sparknlp_jsl.annotator import (
            GenericSVMClassifierApproach,
            GenericSVMClassifierModel,
            GenericLogRegClassifierApproach,
            GenericClassifierModel,
            AssertionLogRegModel,
            AssertionDLModel,
            DeIdentificationModel,
            DocumentLogRegClassifierModel,
            RelationExtractionModel,
            RelationExtractionDLModel,
            ChunkMergeModel,
            SentenceEntityResolverModel,
            ChunkMapperModel,
            BertSentenceChunkEmbeddings,
            ChunkKeyPhraseExtraction,
            NerDisambiguatorModel,
            EntityChunkEmbeddings,
            ZeroShotRelationExtractionModel,
            TFGraphBuilder,
            ChunkConverter,
            ChunkFilterer,
            NerConverterInternal,
            NerChunker,
            AssertionFilterer,
            AnnotationMerger,
            RENerChunksFilter,
            ChunkSentenceSplitter,
            DrugNormalizer,
            ChunkMapperFilterer,
            DateNormalizer,
            GenericClassifierModel,
            ReIdentification,
            Replacer,
            AssertionChunkConverter,
            AssertionLogRegApproach,
            AssertionDLApproach,
            DeIdentification,
            DocumentLogRegClassifierApproach,
            RelationExtractionApproach,
            ChunkMergeApproach,
            SentenceEntityResolverApproach,
            ChunkMapperApproach,
            NerDisambiguator,
            ContextualParserApproach,
            ContextualParserModel,
            GenericClassifierApproach,
            Router,
            NerQuestionGenerator,
            DocumentHashCoder,
            DocMapperModel,
            DocMapperApproach,
            NameChunkObfuscatorApproach,
            NameChunkObfuscator,
            DocumentMLClassifierApproach,
            DocumentMLClassifierModel,
            Resolution2Chunk,
            MedicalQuestionAnswering,
            NerTemplateRenderModel,
            AverageEmbeddings,
            Doc2ChunkInternal,
            Chunk2Token,
            ExtractiveSummarization,
            ChunkFiltererApproach,
            FewShotClassifierModel,
            FewShotClassifierApproach,
            InternalDocumentSplitter,
            Text2SQL,
            IOBTagger,
            DocumentFiltererByClassifier,
        )
        from sparknlp_jsl.structured_deidentification import StructuredDeidentification
        from sparknlp_jsl.modelTracer import ModelTracer
        from sparknlp_jsl import training_log_parser, Deid
        from sparknlp_jsl.training_log_parser import ner_log_parser

        from sparknlp_jsl.base import FeaturesAssembler

        from sparknlp_jsl.annotator.resolution.resolver_merger import ResolverMerger

        from sparknlp_jsl.annotator import (
            MedicalDistilBertForSequenceClassification as DistilBertForSequenceClassification,
            MedicalBertForSequenceClassification as BertForSequenceClassification,
            MedicalBertForTokenClassifier as BertForTokenClassification,
            MedicalNerModel as NerModel,
            MedicalNerApproach as NerApproach,
            MedicalQuestionAnswering as QuestionAnswering,
            MedicalTextGenerator as TextGenerator,
            MedicalSummarizer as Summarizer,
            InternalDocumentSplitter as DocumentSplitter,
        )
        from sparknlp_jsl.compatibility import Compatibility
        from sparknlp_jsl.pretrained import InternalResourceDownloader
        from sparknlp_jsl.eval import (
            NerDLMetrics,
            NerDLEvaluation,
            SymSpellEvaluation,
            POSEvaluation,
            NerCrfEvaluation,
            NorvigSpellEvaluation,
        )

except Exception as err:
    log_broken_lib(Software.spark_hc)
    print(f"Error Message : {err}")
    print(f"Error Trace: {traceback.format_exc()}")
    print("Performing reverse compatibility import for medical module")
    reverse_compatibility_import(__file__, globals())

if try_import_lib("sparknlp_jsl") and try_import_lib("sparknlp"):
    if not Software.spark_hc.check_installed_correct_version() and not warning_logged:
        warning_logged = True
        import sparknlp_jsl

        log_outdated_lib(Software.spark_hc, sparknlp_jsl.version())
